// Copyright 1996, 1997, 1998, 1999 by Jon Dart.  All Rights Reserved.

// Console version of Arasan, works with WinBoard

#include "stdafx.h"
#include "board.h"
#include "movegen.h"
#include "search.h"
#include "options.h"
#include "movearr.h"
#include "notation.h"
#include "timectrl.h"
#include "scoring.h"
#include "chessio.h"
#include "bitmap.h"
#include "bearing.h"
#include "globals.h"
#include "log.h"
#include "util.h"
#include "calctime.h"
#include "eco.h"
#include "strpair.h"
extern "C"
{
#include <string.h>
#include <ctype.h>
};
#include "book.h"
#include <fstream.h>
#include <iostream.h>
#include <strstrea.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <sys/timeb.h>

#define LOG_FILE "games.pgn"

static BOOL verbose = FALSE;
static Search *searcher = NULL;

static Move last_move;
static char last_move_image[80];
static char log_file_name[255];
static int32 time_left = 0;
static int32 opp_time = 0;
static int minutes = 5;
static int incr = 0;
static BOOL computer_plays_white = FALSE;
static char *opponent_name = NULL;
static BOOL ics = FALSE;
static BOOL forceMode = FALSE;
static int moves = 40; /* moves in a time control bloc */
static Board *ponder_board = NULL, *main_board = NULL;
static BOOL ponder_move_ok = FALSE;
static Move predicted_move;
static int pondering = 0;
static Move ponder_move;
static Search::Statistics stats, last_stats;
static int time_target = 0;
static int num_pending = 0;
static int computer_rating = 0;
static int opponent_rating = 0;
#define MAX_PENDING 10
// stack of pending commands
static char *pending[MAX_PENDING];
static BOOL trace = FALSE;
static BOOL easy = FALSE;
static BOOL game_end = FALSE;
static int32 last_score = Constants::BIG, last_time = 0;
static Book_Info book_inf;
static ECO *ecoCoder = NULL;

static Time_Info ti;

// forward declaration
static BOOL do_command(char *buf, Board &board);

static BOOL is_move(char *buf)
{
   if (isalpha(*buf) && isdigit(buf[1]) &&
       isalpha(buf[2]) && isdigit(buf[3]))
      return TRUE;
   else if (strcmp(buf,"o-o")==0 || strcmp(buf,"O-O")==0 ||
            strcmp(buf,"o-o-o")==0 || strcmp(buf,"O-O-O") ==0)
      return TRUE;
   else
      return FALSE;
}

static Move text_to_move(const Board &board,char *buf)
{
   Move rmove;
   if (strcmp(buf,"o-o") == 0 ||
       strcmp(buf,"o-o-o") == 0 ||
       strcmp(buf,"O-O") == 0 ||
       strcmp(buf,"O-O-O") == 0)
   {
      char *p =  buf;
      while (*p)
      {
         *p = toupper(*p); p++;
      }
      rmove = Notation::MoveValue(board,buf,board.Side());
   }
   else
   {
      Square start = SquareValue(buf);
      Square dest = SquareValue(buf+2);
      if (buf[4])
      {
         PieceType promotion = 
         PieceCharValue(toupper(buf[4]));
         rmove = CreateMove(board,start,dest,
                            promotion);
      }
      else
         rmove = CreateMove(board,start,dest,InvalidPiece);
   }
   return rmove;
}

static BOOL accept_draw(Board &board)
{
   if (trace)
      cout << "(debug) in accept_draw" << endl;
   // Code to handle draw offers.  This is somewhat experimental.
   int rating_diff = opponent_rating - computer_rating;
   // ignore draw if we have just started searching
   if (last_score == Constants::BIG)
      return FALSE;
   // If it's a 0-increment game and the opponent has < 1 minute,
   // and we have more time, decline
   if (incr == 0 && opp_time < 6000 && time_left > opp_time)
      return FALSE;
   // See if we do not have enough material to mate
   const ColorType side = computer_plays_white ? White : Black;
   const Material &ourmat = board.getMaterial(side);
   const Material &oppmat = board.getMaterial(OppositeColor(side));
   if (ourmat.no_pawns() && ourmat.value() <= KING_VALUE + BISHOP_VALUE)
   {
      // We don't have mating material
      return TRUE;
   }
   // accept a draw in pawnless endings with even material, unless
   // our score is way positive
   if (ourmat.no_pawns() && oppmat.no_pawns() &&
       (Util::Abs(ourmat.value() - oppmat.value()) < 32) &&
       last_score < (int)Constants::PawnValue)
   {
      return TRUE;
   }
   ColorType tmp = board.Side();
   board.Set_Side(side);
   int draw_score = searcher->draw_score(board);
   board.Set_Side(tmp);
   int threshold;
   threshold = Constants::PawnValue/2;
   if (rating_diff > 50)
   {
      threshold += Constants::PawnValue*(rating_diff-50)/200;
   }
   else if (rating_diff < -250)
      threshold = Constants::PawnValue*(rating_diff+100)/150;
   if (trace)
   {
      cout << "(debug) rating_diff = " << rating_diff << endl;
      cout << "(debug) draw_score = " << draw_score << endl;
      cout << "(debug) last_score = " << last_score << endl;
      cout << "(debug) threshold = " << threshold << endl;
   }
   return (draw_score > Constants::PawnValue/2 &&
           last_score < threshold);
}

static void save_game()
{
   if (game_moves->num_moves() == 0)
      return;
   ofstream ofs(log_file_name,ios::out | ios::app);
   if (ofs.good())
   {
      CPtrArray headers;
      CString opening_name, eco;
      if (ecoCoder)
      {
        ecoCoder->classify(*game_moves,eco,opening_name);
        headers.Add(new StringPair("ECO",eco));
      }
      static char timec[50];
	  static char crating[15];
	  static char orating[15];
	  sprintf(crating,"%d",computer_rating);
	  sprintf(orating,"%d",opponent_rating);

      if (computer_plays_white)
      {
        headers.Add(new StringPair("Black",
			opponent_name ? opponent_name : "?"));
        if (computer_rating)
           headers.Add(new StringPair("WhiteElo",crating));
        if (opponent_rating)
           headers.Add(new StringPair("BlackElo",orating));
      }
      else
      {
        headers.Add(new StringPair("White",
			opponent_name ? opponent_name : "?"));
        if (opponent_rating)
           headers.Add(new StringPair("WhiteElo",orating));
        if (computer_rating)
           headers.Add(new StringPair("BlackElo",crating));
      }
      sprintf(timec,"%d+%d",minutes*60,incr);
      headers.Add(new StringPair("TimeControl",timec));
      ChessIO::store_pgn(ofs, *theLog,
                         computer_plays_white ? White : Black,
                         headers);
      ofs.close();
   }
   else
   {
      cerr << "error opening logfile" << endl;
      MessageBox(NULL,"error opening log","",MB_OK);
   }
}

static int calc_extra_time()
{
   // Determine if we are allowed to use any time beyond
   // the normal time target.
   if (time_left < 1500)
      return 0; // no extra time
   else if (time_target <= 100)
      return 100; // lightning game
   else if ((incr == 0 && time_left > time_target*10) ||
            (incr > 0 && time_left > time_target*6))
      return time_target;
   else
      return 0;
}

// handle ^C interrupts (sent by winboard)
static void __cdecl sigHandler(int)
{
   cout << "got ^C" << endl;
   if (searcher)
   {
      searcher->terminate_now();
   }
   signal(SIGINT,sigHandler);
}

static void move_image(const Board &board, Move m, char *buf)
{
   // Produce a text string representing the move, in a form
   // Xboard understands.
   if (TypeOfMove(m) == KCastle)
   {
      strcpy(buf,"o-o");
   }
   else if (TypeOfMove(m) == QCastle)
   {
      strcpy(buf,"o-o-o");
   }
   else
   {
      *buf++ = FileImage(StartSquare(m));
      *buf++ = RankImage(StartSquare(m));
      *buf++ = FileImage(DestSquare(m));
      *buf++ = RankImage(DestSquare(m));
      if (TypeOfMove(m) == Promotion)
      {
         // N.b. ICS requires lower case.
         *buf++ = tolower(PieceImage(PromoteTo(m)));
      }
      *buf = '\0';
   }                  
}

static void post_output(const Search::Statistics &stats)
{
   last_score = stats.value;
   last_time = stats.elapsed_time;
   if (verbose)
   {
      // "post" output for Winboard
      printf("%2d%c %6d %6d %8d %s\n",
             stats.depth,' ',
             (100*stats.value)/64,
             stats.elapsed_time,
             stats.num_nodes,
             stats.best_line_image);
      fflush(stdout);
   }
}

// handle editing the chessboard
static void edit_board(istream &fin, Board &board)
{
   for (int i = 0; i < 64; i++)
      board.Set_Contents(EmptyPiece,i);
   static char buf[80];
   static const char pieces[]="PNBRQKpnbrqk";
   ColorType side = White;
   while (fin.getline(buf,80))
   {
      if (strcmp(buf,"white") == 0)
         side = White;
      else if (strcmp(buf,"black")==0)
         side = Black;
      else if (*buf == '#')
      {
         for (int i = 0; i < 64; i++)
            board.Set_Contents(EmptyPiece,i);
      }
      else if (strcmp(buf,"c")==0)
         side = OppositeColor(side);
      else if (strcmp(buf,".")==0)
         break;
//    else if (!strcmp(buf,"d")) {
//      DisplayChessBoard(stdout,search);
//    }
      else
      {
         if (strchr(pieces,buf[0]))
         {
            char c = tolower(buf[0]);
            Piece p;
            switch (c)
            {
            case 'p':
               p = MakePiece(Pawn,side);
               break;
            case 'n':
               p = MakePiece(Knight,side);
               break;
            case 'b':
               p = MakePiece(Bishop,side);
               break;
            case 'r':
               p = MakePiece(Rook,side);
               break;
            case 'q':
               p = MakePiece(Queen,side);
               break;
            case 'k':
               p = MakePiece(King,side);
               break;
            default:
               break;
            }

            if (p != InvalidPiece)
            {
               Square sq = SquareValue(buf+1);
               if (!IsInvalid(sq))
                  board.Set_Contents(p,sq);
            }
         }
      }
   }
   board.set_secondary_vars();
   // edit doesn't set the castle status, so try to deduce it 
   // from the piece positions
   if (board.KingPos(White) == (Square)60)
   {
      if (board[56] == MakePiece(Rook,White) &&
          board[63] == MakePiece(Rook,White))
         board.Set_CastleStatus(Board::CanCastleEitherSide,White);
      else if (board[56] == MakePiece(Rook,White))
         board.Set_CastleStatus(Board::CanCastleQSide,White);
      else if (board[63] == MakePiece(Rook,White))
         board.Set_CastleStatus(Board::CanCastleKSide,White);
   }
   else
      board.Set_CastleStatus(Board::CantCastleEitherSide,White);
   if (board.KingPos(Black) == (Square)4)
   {
      if (board[0] == MakePiece(Rook,Black) &&
          board[7] == MakePiece(Rook,Black))
         board.Set_CastleStatus(Board::CanCastleEitherSide,Black);
      else if (board[0] == MakePiece(Rook,Black))
         board.Set_CastleStatus(Board::CanCastleQSide,Black);
      else if (board[7] == MakePiece(Rook,Black))
         board.Set_CastleStatus(Board::CanCastleKSide,Black);
   }
   else
      board.Set_CastleStatus(Board::CantCastleEitherSide,Black);
}  

static void ponder(Board &board, Move move, Move predicted_reply)
{
   ponder_move_ok = FALSE;
   ponder_move = NullMove;
   if (!IsNull(move) && !IsNull(predicted_reply))
   {
      // start pondering
      predicted_move = predicted_reply;
      if (ponder_board) delete ponder_board;
      ponder_board = new Board(board);
      // Make the move we predicted
      ponder_board->DoMove(predicted_reply);
      // Start a search for the reply
      stats.clear();
      book_inf.clear();
      ponder_move = opening_book->book_move(board,Balanced,book_inf);
      if (IsNull(ponder_move))
      {
         pondering++;
         if (trace)
            cout << "(debug) starting to ponder" << endl;
         ponder_move =
         searcher->find_best_move(NULL,*ponder_board,
                                  Search::Fixed_Time,
                                  99999 /* time limit (999 seconds) */,
                                  0, /* extra time allowed */
                                  99 /* ply limit */,
                                  TRUE /* background */,
                                  stats,
                                  (trace) ? Search::Trace : Search::Silent);
         pondering--;
         if (trace)
            cout << "(debug) done pondering" << endl;
      }
      else // we got a book move
         ponder_move = NullMove;
   }
}

static Move search(Search &searcher, Board &board, Search::Statistics &stats)
{
   Move pred = NullMove;
   last_stats.clear();
   last_score = Constants::BIG; last_time = 0;
   book_inf.clear();
   Move move = opening_book->book_move(board,Balanced,book_inf);
   if (IsNull(move))
   {
      // no book move
      time_target = calc_time_limit(moves, minutes,incr,time_left,opp_time);
      if (ics && time_target >= 300)
         cout << "whisper time target = " << (time_target+50)/100
           << " seconds" << endl;

      Search_Limit limit;
      limit.centiseconds = time_target;
      stats.clear(); 
      if (trace)
         cout << "(debug) entering search" << endl;
      move = searcher.find_best_move(NULL, board, 
                                     Search::Time_Limit, time_target, 
                                     calc_extra_time(),
                                     9999, FALSE,
                                     stats, 
                                     Search::Whisper);
      if (trace)
         cout << "(debug) search done " << endl;
      last_stats = stats;
      post_output(stats);
   }
   else
   {
      cout << "whisper book move" << endl;
   }
   return move;
}

static void send_move(Board &board, Move &move, Search::Statistics
                      &stats)
{
   last_move = move;
   last_stats = stats;
   if (!IsNull(move))
   {
      int rep_count = board.rep_count();
      if (rep_count >= 2)
      {
         // It is possible the opponent has not claimed the
         // draw by repetition. If it's a draw before we even
         // move, don't make a move, just claim the draw
         cout << "1/2-1/2 {Repetition}" << endl;
         if (trace) cout << "(debug) repetition draw" << endl;
         theLog->setResult("1/2-1/2 {repetition}");
         return;
      }
 
      Notation::Image(board,last_move,last_move_image);
      theLog->add_move(board,last_move,last_move_image,&last_stats,&book_inf,TRUE);
      opening_book->learn(board,rep_count);
      char movebuf[20];
      move_image(board,last_move,movebuf);
      cout << 1+game_moves->num_moves()/2 << ". ... ";
      cout << movebuf << endl;
      // for Gnuchess compatibility
      cout << "My move is: " << movebuf << endl;
      board.DoMove(last_move);
      game_moves->add_move(board,last_move);
   }
   else
      cout << "(debug) warning : move is null" << endl;
   if (ics && time_target >= 300)
   {
      char val[20];
      sprintf(val,"%6.2f",last_stats.value/64.0);
      static char statbuf[80];

      if (last_stats.num_moves > 0)
      {
         if (last_stats.elapsed_time > 0)
         {
            int nps = 
              (int)((last_stats.num_nodes*100)/last_stats.elapsed_time);
            sprintf(statbuf,"time : %d.%02d seconds  score: %6.2f  depth: %d  nps : %d",
                 last_stats.elapsed_time/100,
                 last_stats.elapsed_time % 100,
                 last_stats.display_value/64.0,
                 last_stats.depth,
                 nps);
         }
         else
            sprintf(statbuf,"time : %d.%02d seconds  score: %6.2f  depth: %d nodes: %d",
                 last_stats.elapsed_time/100,
                 last_stats.elapsed_time % 100,
                 last_stats.display_value/64.0,
                 last_stats.depth,
                 last_stats.num_nodes);
         
         if (ics) cout << "tellics whisper " << statbuf << endl;
      }
      if (last_stats.best_line_image[0] != '\0')
      {
         char *p = strchr(last_stats.best_line_image,' ');
         if (p && *(p+1) != '\0')
            cout << "whisper predicting: " << p+1 << endl;
      }
   }
   if (ics && last_stats.value > Constants::BIG-100 &&
       last_stats.value < Constants::BIG-1)
   {
      int movesToMate = (Constants::BIG-last_stats.value+1)/2;
      cout << "whisper Arasan will mate in " << movesToMate <<
      (movesToMate == 1 ? "move." :  " moves.") << endl;
   }
   if (last_stats.state == Search::Stalemate)
   {
      cout << "1/2-1/2 {Stalemate}" << endl;
      theLog->setResult("1/2-1/2 {stalemate}");
   }
   else if (board.move_count >= 100)
   {
      cout << "1/2-1/2 {50 move rule}" << endl;
      theLog->setResult("1/2-1/2 {50 move draw}");
      if (trace) cout << "(debug) 50 move draw" << endl;
   }
   else if (Scoring::material_draw(board))
   {
      cout << "1/2-1/2 {Insufficient material}" << endl;
      theLog->setResult("1/2-1/2 {insufficient material}");
      if (trace) cout << "(debug) material draw" << endl;
   }
   else if (Scoring::repetition_draw(board))
   {
      cout << "1/2-1/2 {Repetition}" << endl;
      if (trace) cout << "(debug) repetition draw" << endl;
      theLog->setResult("1/2-1/2 {repetition}");
   }
   else if (stats.state == Search::Draw)
   {
      // We can get here if our opponent does not claim the
      // repetition draw.
      cout << "(debug) opponent didn't claim repetition?" << endl;
   }
   else if (last_stats.value == 9999)
   {
      if (computer_plays_white)
      {
         theLog->setResult("1-0");
         cout << "1-0 {White mates}" << endl;
      }
      else
      {
         theLog->setResult("0-1");
         cout << "0-1 {Black mates}" << endl;
      }
   }
   else if (last_stats.state == Search::Checkmate)
   {
      if (computer_plays_white)
      {
         theLog->setResult("0-1");
         cout << "0-1 {Black mates}" << endl;
      }
      else
      {
         theLog->setResult("1-0");
         cout << "1-0 {White mates}" << endl;
      }
   }
   else if (last_stats.state == Search::Resigns)
   {
      // Don't resign a zero-increment game if the opponent is short
      // on time
      if (!(incr == 0 && opp_time < 2000))
      {
         // Winboard passes the resign command to ICS, but ignores it
         // itself.
         if (computer_plays_white)
         {
            theLog->setResult("0-1 {White resigns}");
            cout << "0-1 {White resigns}" << endl;
         }
         else
         {
            theLog->setResult("1-0 {Black resigns}");
            cout << "1-0 {Black resigns}" << endl;
         }
      }
   }
}    

static void undo( Board &board)
{
   board.UndoMove((*theLog)[theLog->current()-1].move(),
                  (*theLog)[theLog->current()-1].state());
   theLog->back_up();
   game_moves->remove_move();
   if (theLog->current())
   {
      last_move = (*theLog)[theLog->current()-1].move();
      strcpy(last_move_image,(*theLog)[theLog->current()-1].image());
   }
   else
   {
      last_move = NullMove;
      *last_move_image = '\0';
   }
}

static void add_pending(char *buf)
{
   if (num_pending < MAX_PENDING)
   {
      pending[num_pending] = strdup(buf);
      ++num_pending;
   }
}

static BOOL do_pending(Board &board)
{
   while (num_pending)
   {
      --num_pending;
      BOOL terminate;
      if (trace)
         cout << "(debug) pending command: " << pending[num_pending] << endl;
      terminate = !do_command(pending[num_pending],board);
      free( pending[num_pending]);
      if (terminate)
      {
         return FALSE;
      }
   }
   return TRUE;
}

// Execute a move made by the opponent or in "force" mode.
static void execute_move(Board &board,Move m)
{
   last_move = m;
   Notation::Image(board,m,last_move_image);
   theLog->add_move(board,m,last_move_image,NULL,NULL,TRUE);

   board.DoMove(m);

   game_moves->add_move(board,m);
}

// Handle a command received while searching.  terminate is set
// TRUE if the search should stop.  Some commands are executed,
// some are placed on the pending stack to be executed after
// the search completes.
void check_command(char *buf, BOOL &terminate)
{
   if (trace)
      cout << "(debug) check_command: " << buf << endl;
   terminate = FALSE;
   if (strcmp(buf,"quit")==0 || strcmp(buf,"end")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"new")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"?") == 0)
   {
      // Winboard 3.6 sends this to terminate a search in progress
      if (trace) cout << "(debug) terminating." << endl;
      terminate = TRUE;
   }
   else if (strcmp(buf,"hint")==0)
   {
      // not yet implemented
   }
   else if (strcmp(buf,"depth")==0)
   {
   }
   else if (strncmp(buf,"level",5)==0)
   {
      sscanf(buf+6,"%ld %ld %ld",&moves,&minutes,&incr);
   }
   else if (strncmp(buf,"time",4)==0)
   {
      // my time left
      sscanf(buf+4,"%ld",&time_left);
   }
   else if (strncmp(buf,"otime",4)==0)
   {
      // opponent's time left
      sscanf(buf+4,"%ld",&opp_time);
   }
   else if (strcmp(buf,"post")==0)
   {
      verbose = TRUE;
   }
   else if (strcmp(buf,"nopost")==0)
   {
      verbose = FALSE;
   }
   else if (strcmp(buf,"savegame")==0)
   {
      
   }
   else if (strcmp(buf,"remove")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"undo")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"resign")==0 || strncmp(buf,"result",6)==0)
   {
      add_pending(buf);
      game_end = TRUE;
      terminate = TRUE;
   }
   else if (strcmp(buf,"draw")==0)
   {
      // "draw" command. Requires winboard 3.6 or higher.
      if (accept_draw(*main_board))
      {
         // Notify opponent. don't assume draw is concluded yet.
         cout << "offer draw" << endl;
      }
      //else
      //   cout << "tellics decline draw" << endl;
      // terminate = TRUE;
   }
   else if (strcmp(buf,"edit")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"go")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"bk")==0)
   {
      
   }
   else if (strcmp(buf,"easy")==0)
   {
      easy = TRUE;
   }
   else if (strcmp(buf,"hard")==0)
   {
      easy = FALSE;
   }
   else if (strcmp(buf,"white")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strcmp(buf,"black")==0)
   {
      add_pending(buf);
      terminate = TRUE;
   }
   else if (strncmp(buf,"name",4)==0)
   {
      // We've received the name of our opponent.
      opponent_name = strdup(buf+5);
   }
   else if (strncmp(buf,"bogus",5)==0)
   {
      
   }
   else if (strncmp(buf,"force",5)==0)
   {
      forceMode = !forceMode;      
      terminate = TRUE;
   }
   else if (strncmp(buf,"rating",6) == 0)
   {
      if (trace) cout << "(debug) rating = " << buf << endl;
      sscanf(buf+7,"%d %d",&computer_rating, &opponent_rating);
   }
   else
   {
      // see if it could be a move
      if (isdigit(*buf))
      {
         char *p = buf;
         while (*p && !isalpha(*p)) ++p;
         if (*p)
         {
            char *q = buf;
            while (*p) *q++=*p++;
         }
      }
      if (trace)
         cout << "(debug) move text = " << buf << endl;
      if (is_move(buf))
      {
         // make the move on the board
         Move rmove = text_to_move(*main_board,buf);
         last_move = rmove;

         if (forceMode)
         {
            add_pending(buf);
            terminate = TRUE;
         }
         else if (!IsNull(predicted_move) &&
                  MovesEqual(predicted_move,last_move))
         {
            // make the move on the board
            execute_move(*main_board,last_move);

            // We predicted the opponent's move, so we need to
            // continue doing the search but adjust the time
            // limit.
            if (ics && time_target >= 300)
               cout << "whisper predicted last move" << endl;
            ponder_move_ok = TRUE;
            int elapsed_time = (GetTickCount()/10)-searcher->get_start_time();
            time_target = calc_time_limit(moves, minutes,incr,time_left,opp_time);
            if (trace)
            {
               cout << "(debug) time_target = " << time_target << endl;
            }
            // Compute how much longer we must search
            int32 new_time;
            if (elapsed_time > time_target)
            {
               new_time = elapsed_time;
            }
            else
            {
               new_time = time_target;
            }
            if (new_time <= 0) new_time = 30;
            // Normally we will move fast if the last move
            // was predicted. But slow down if we are 
            // way ahead of the opponent or have lots of time.
            /**if (minutes > 0 && time_left > 200*incr && 
                new_time - elapsed_time < incr*100 &&
                (time_left > minutes*6000 ||
                 time_left > (opp_time*3)/2))
            {                 
               new_time = elapsed_time + incr*100;
            }
            if (trace)
               cout << "(debug) new search limit = " << new_time <<endl;**/
            searcher->set_time_limit(new_time);
            searcher->set_talk_level(Search::Whisper);
            searcher->set_background(FALSE);
            terminate = FALSE;
         }
         else
         {
            // We can't use the results of pondering because we
            // did not predict the opponent's move.  Stop the
            // search and then execute the move.
            ponder_move_ok = FALSE;
            add_pending(buf);
            terminate = TRUE;
         }
      }
   }
}

// Execute a command, return FALSE if program should terminate.
static BOOL do_command(char *buf, Board &board)
{
   if (trace)
      cout << "(debug) do_command: " << buf << endl;
   if (strcmp(buf,"quit")==0 || strcmp(buf,"end")==0)
   {
      return FALSE;
   }
   else if (strcmp(buf,"new")==0)
   {
      save_game();  
      board.Reset();
      theLog->reset();
      theLog->write_header();
      computer_plays_white = FALSE;
      forceMode = FALSE;
      game_end = FALSE;
      stats.clear();
      last_stats.clear();
      last_move = NullMove;
      *last_move_image = '\0';
      delete game_moves;
      game_moves = new Move_Array();
      delete searcher;
      predicted_move = NullMove;
      pondering = 0;
//      delete opening_book;
//      opening_book = new Book(); 
      ecoCoder = new ECO();
      searcher = new Search();
      searcher->register_command_function(check_command);
      searcher->register_post_function(post_output);
      if (ics)
        cout << "kib Hello from Arasan version " << Arasan_Version << endl;
   }
   else if (strcmp(buf,"hint")==0)
   {
      // not yet implemented
   }
   else if (strcmp(buf,"bk")==0)
   {
      Move moves[20];
      unsigned count = opening_book->book_moves(*main_board,moves,20);
      if (count == 0)
      {
         cout << '\t' << "No book moves for this position." << endl
         << endl;
      }
      else
      {
         cout << " book moves:" << endl;
         for (unsigned i = 0; i<count; i++)
         {
            char result[64];
            Notation::Image(*main_board,moves[i],result);
            cout << '\t' << result << endl;
         }
         cout << endl;
      }
   }
   else if (strcmp(buf,"depth")==0)
   {
      
   }
   else if (strncmp(buf,"level",5)==0)
   {
      sscanf(buf+6,"%ld %ld %ld",&moves,&minutes,&incr);
//         cout << "level command: " << buf;
   }
   else if (strncmp(buf,"time",4)==0)
   {
      // my time left
      sscanf(buf+4,"%ld",&time_left);
   }
   else if (strncmp(buf,"otime",4)==0)
   {
      sscanf(buf+4,"%ld",&opp_time);
   }
   else if (strcmp(buf,"post")==0)
   {
      verbose = TRUE;
   }
   else if (strcmp(buf,"nopost")==0)
   {
      verbose = FALSE;
   }
   else if (strncmp(buf,"result",6)==0)
   {
      // Game has ended
      theLog->setResult(buf+7);
      save_game();
      delete game_moves;
      game_moves = new Move_Array();
   }
   else if (strcmp(buf,"savegame")==0)
   {
      
   }
   else if (strcmp(buf,"remove")==0)
   {
      undo(board);
      undo(board);
   }
   else if (strcmp(buf,"undo")==0)
   {
      undo(board);
   }
   else if (strcmp(buf,"resign")==0)
   {
      // our opponent has resigned
      if (computer_plays_white)
         theLog->setResult("1-0");
      else
         theLog->setResult("0-1");
   }
   else if (strcmp(buf,"draw")==0)
   {
      // "draw" command. Requires winboard 3.6 or higher.
      if (accept_draw(board))
      {
         // Notify opponent. don't assume draw is concluded yet.
         cout << "offer draw" << endl;
      }
      //else
      //   cout << "tellics decline draw" << endl;
   }
   else if (strcmp(buf,"edit")==0)
   {
      edit_board(cin,board);
   }
   else if (strcmp(buf,"go")==0)
   {
      cout << "1. go" << endl;
      Move reply = search(*searcher,board,stats);
      send_move(board,reply,stats);
   }
   else if (strcmp(buf,"bk")==0)
   {
      
   }
   else if (strcmp(buf,"easy")==0)
   {
      easy = TRUE;
   }
   else if (strcmp(buf,"hard")==0)
   {
      easy = FALSE;
   }
   else if (strcmp(buf,"white")==0)
   {
      computer_plays_white = TRUE;
      forceMode = FALSE;
   }
   else if (strcmp(buf,"black")==0)
   {
      computer_plays_white = FALSE;
      forceMode = FALSE;
   }
   else if (strncmp(buf,"name",4)==0)
   {
      // We've received the name of our opponent.
      opponent_name = strdup(buf+5);
   }
   else if (strncmp(buf,"bogus",5)==0)
   {
      
   }
   else if (strncmp(buf,"force",5)==0)
   {
      forceMode = !forceMode;      
   }
   else if (strncmp(buf,"rating",6) == 0)
   {
      if (trace) cout << "(debug) rating = " << buf << endl;
      sscanf(buf+7,"%d %d",&computer_rating, &opponent_rating);
   }
   else
   {
      // see if it could be a move
      if (isdigit(*buf))
      {
         char *p = buf;
         while (*p && !isalpha(*p)) ++p;
         if (*p)
         {
            char *q = buf;
            while (*p) *q++=*p++;
         }
      }
//          cout << "move text = " << buf << endl;
      if (is_move(buf))
      {
         if (trace)
            cout << "(debug) got move: " << buf << endl;
         // assume move
         cout << game_moves->num_moves() << ". " << buf << endl;

         Move rmove = text_to_move(board,buf);
         if (!IsNull(rmove))
         {
            // make the move on the board
            execute_move(board,rmove);
            Move reply;
            if (!forceMode)
            {
               reply = search(*searcher,board,stats);
               if (!game_end)
                 send_move(board,reply,stats);
            }
            if (!forceMode && !easy && time_target >= 300 /* 3 seconds */)
            {
               for (;;)
               {
                  ponder(board, reply, stats.best_line[1]);
                  if (ponder_move_ok && !IsNull(ponder_move) && !game_end)
                  {
                     // we got a reply from pondering
                     if (trace)
                        cout << "(debug) sending ponder move" << endl;
                     send_move(board,ponder_move,stats);
                     post_output(stats);
                     reply = ponder_move;
                  }
                  else // no ponder move
                  {
                     return TRUE;
                  }
               }
            }
         }
      }
   }
   return TRUE;
}

int __cdecl main(int argc, char **argv)
{
   signal(SIGINT, sigHandler);

   init_globals(argv[0],TRUE,TRUE);
   atexit(cleanup_globals);

   Board board;
   main_board = &board;
   Search_Limit limit;
   Search_Type type = Fixed_Ply;
   limit.max_ply = 2;
   int arg = 1;

   if (argc > 1)
   {
      while (arg < argc && *(argv[arg]) == '-')
      {
         char c = *(argv[arg]+1);
         switch (c)
         {
         case 'i':
            if (strcmp(argv[arg]+1,"ics")==0)
               ics = TRUE;
            else
            {
               cerr << "Warning: unknown option: " << argv[arg]+1 <<
               endl;
            }
            break;
         case 'H': 
            ++arg;
            SearchOps ops;
            global_options->get_search_ops(ops);
            ops.hash_table_size = atol(argv[arg]);
            global_options->update_search_ops(ops);
            break;      
         case 't':
            //++arg;
            trace = TRUE;
            break;
         default:
            cerr << "Warning: unknown option: " << argv[arg]+1 <<
            endl;
            break;
         }
         ++arg;
      }
   }
   if (arg < argc)
   {
      cout << "loading " << argv[arg] << endl;
      ifstream pos_file( argv[arg], ios::in | ios::nocreate);
      if (pos_file.good())
      {
         pos_file >> board;
      }
      else
      {
         cout << "file not found: " << argv[arg] << endl;
         return -1;
      }
   }
   *last_move_image = '\0';

   strcpy(log_file_name,programPath);
   char *p = strrchr(log_file_name,'\\');
   if (p)
      strcpy(p+1,LOG_FILE);
   else
      strcpy(log_file_name,LOG_FILE);

   searcher = new Search();
   searcher->register_command_function(check_command);
   searcher->register_post_function(post_output);

   // Show a message on the console
   cout << "Arasan " Arasan_Version << " Copyright 1997-2000 by Jon Dart" << endl;

   // Must use unbuffered console
   setbuf(stdin,0);
   int nextc;
   
   // Read stdin until we get a quit command
   while ((nextc = getc(stdin)) != EOF)
   {
      if (nextc == '\n')
      {
         cmd_buf[cmd_buf_count] = '\0';
         cmd_buf_count = '\0';
//       cout << "received " << cmd_buf << endl;     
         if (!do_command(cmd_buf,board))
            break;
         if (!do_pending(board))
            break;
      }
      else
      {
         if (cmd_buf_count >= 255)
         {
            cout << "(debug) warning: command buffer full" << endl;
            cmd_buf_count = 0;
         }
         cmd_buf[cmd_buf_count++] = nextc;
      }
   }

   // handle termination.
   if (opp_time <= 0)
   {
      if (computer_plays_white)
         theLog->setResult("1-0");
      else
         theLog->setResult("0-1");
   }
   else if (time_left <= 0)
   {
      if (computer_plays_white)
         theLog->setResult("0-1");
      else
         theLog->setResult("1-0");
   }
   save_game();
   searcher->terminate_now();
   delete searcher;

   cout << "quit" << endl;
   Sleep(2000);
   return 0;
}
