// Copyright 1994-1998 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "board.h"
#include "constant.h"
#include "bhash.h"
#include "util.h"
#include "bearing.h" 
#include "debug.h"
#include "globals.h"
#include "boardio.h"
#include "boardsta.h"
#include <strstrea.h>
#include <ctype.h>
#include <memory.h>

enum
{
   BoardSize = 64
};
enum
{
   RepListSize = 500
};

void Board::set_secondary_vars()
{
   int index[2], i;

   index[Black] = index[White] = 0;

   my_Material[White].clear();
   my_Material[Black].clear();
   pawn_bits[White].clear();
   pawn_bits[Black].clear();
   knight_bits[White].clear();
   knight_bits[Black].clear();
   bishop_bits[White].clear();
   bishop_bits[Black].clear();
   rook_bits[White].clear();
   rook_bits[Black].clear();
   rooksqueens[White].clear();
   rooksqueens[Black].clear();
   bishopsqueens[White].clear();
   bishopsqueens[Black].clear();
   queen_bits[White].clear();
   queen_bits[Black].clear();
   king_bits[White].clear();
   king_bits[Black].clear();
   occupied[White].clear();
   occupied[Black].clear();
   all_occupied.clear();
   occupiedR90.clear();
   occupiedDa1.clear();
   occupiedDa8.clear();
   for (i=0;i<BoardSize;i++)
   {
      Square sq(i);
      if (my_Contents[sq] != EmptyPiece)
      {
         const Piece piece = my_Contents[sq];
         const ColorType color = PieceColor(piece);
         occupied[color].set(sq);
         all_occupied.set(sq);
         occupiedR90.setR90(sq);
         occupiedDa1.setDa1(sq);
         occupiedDa8.setDa8(sq);
         my_Material[color].add_piece(TypeOfPiece(piece));
         switch (TypeOfPiece(piece))
         {
         case King:
            my_KingPos[color] = sq;
            king_bits[color].set(sq);
            break;
         case Pawn:
            pawn_bits[color].set(sq);
            break;
         case Knight:
            knight_bits[color].set(sq);
            break;
         case Bishop:
            bishop_bits[color].set(sq);
            bishopsqueens[color].set(sq);
            break;
         case Rook:
            rook_bits[color].set(sq);
            rooksqueens[color].set(sq);
            break;
         case Queen:
            queen_bits[color].set(sq);
            rooksqueens[color].set(sq);
            bishopsqueens[color].set(sq);
            break;
         default:
            break;
         }
      }
   }
   my_HashCode = Board_Hash::HashCode(*this);
   m_pawn_hash = Board_Hash::pawn_hash(*this);
   move_count = 0;
}

void Board::Reset()
{
   static PieceType pieces[] =
   {
      Rook,
      Knight,
      Bishop,
      Queen,
      King,
      Bishop,
      Knight,
      Rook
   };

   my_Side = White;
   my_CheckStatus = Unknown;
   for (int i=0;i<BoardSize;i++)
   {
      const Square sq(i);
      if (Rank(sq,White) == 1)
         my_Contents[sq] = MakeWhitePiece( pieces[File(sq)-1]);
      else if (Rank(sq,Black) == 1)
         my_Contents[sq] = MakeBlackPiece( pieces[File(sq)-1]);
      else if (Rank(sq,White) == 2)
         my_Contents[sq] = whitePawn;
      else if (Rank(sq,Black) == 2)
         my_Contents[sq] = blackPawn;
      else
         my_Contents[sq] = EmptyPiece;
   }
   my_EnPassantSq[White] = InvalidSquare;
   my_EnPassantSq[Black] = InvalidSquare;
   my_CastleStatus[White] = my_CastleStatus[Black] = CanCastleEitherSide;
   set_secondary_vars();
   if (rep_list)
      delete [] rep_list;
   rep_list = rep_list_head = new hash_t[RepListSize];
   move_count = 0;
}

Board::Board()
: rep_list(NULL)
{
   Reset();
}

Board::Board(const Board &b)
{
   my_Side = b.my_Side;
   my_CheckStatus = b.my_CheckStatus;
   for (int i=0;i<BoardSize;i++)
   {
      my_Contents[i] = b[i];
   }
   my_EnPassantSq[White] = b.my_EnPassantSq[White];
   my_EnPassantSq[Black] = b.my_EnPassantSq[Black];
   my_CastleStatus[White] = b.my_CastleStatus[White];
   my_CastleStatus[Black] = b.my_CastleStatus[Black];
   set_secondary_vars();

   // Copy the repetition table
   move_count = b.move_count;
   rep_list = new hash_t[RepListSize];
   int rep_entries = b.rep_list_head - b.rep_list;
   for (i = 0; i < rep_entries; i++)
      rep_list[i] = b.rep_list[i];
   rep_list_head = rep_list + rep_entries;
}

Board &Board::operator =(const Board &b)
{
   if (&b != this)
   {
      my_Side = b.my_Side;
      my_CheckStatus = b.my_CheckStatus;
      for (int i=0;i<BoardSize;i++)
      {
         my_Contents[i] = b[i];
      }
      my_EnPassantSq[White] = b.my_EnPassantSq[White];
      my_EnPassantSq[Black] = b.my_EnPassantSq[Black];
      my_CastleStatus[White] = b.my_CastleStatus[White];
      my_CastleStatus[Black] = b.my_CastleStatus[Black];
      set_secondary_vars();

      // Copy the repetition table
      move_count = b.move_count;
      delete [] rep_list;
      rep_list = new hash_t[RepListSize];
      int rep_entries = b.rep_list_head - b.rep_list;
      for (i = 0; i < rep_entries; i++)
         rep_list[i] = b.rep_list[i];
      rep_list_head = rep_list + rep_entries;
   }
   return *this;
}

Board::~Board()
{
   delete [] rep_list;
}

#ifdef _DEBUG
const Piece &Board::operator[]( const Square sq ) const
{
   ASSERT(OnBoard(sq));
   return my_Contents[sq];
}
#endif

void Board::UpdateCastleStatus( const ColorType side, 
                                const Square sq )
// after a move of or capture of the rook on 'sq', update castle status
// for 'side'
{
   CastleType cs = CastleStatus(side);
   if ((cs != CastledKSide) && (cs != CastledQSide))
   {
      if (File(sq) == 1) // Queen Rook moved or captured
      {
         if (cs == CanCastleEitherSide)
            my_CastleStatus[side] = CanCastleKSide;
         else if (cs == CanCastleQSide)
            my_CastleStatus[side] = CantCastleEitherSide;
         if (side == White)
         {
           my_HashCode ^= w_castle_status[(int)cs];
           my_HashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         }
         else
         {
           my_HashCode ^= b_castle_status[(int)cs];
           my_HashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         }
      }
      else if (File(sq) == 8) // King Rook moved or captured
      {
         if (cs == CanCastleEitherSide)
            my_CastleStatus[side] = CanCastleQSide;
         else if (cs == CanCastleKSide)
            my_CastleStatus[side] = CantCastleEitherSide;
         if (side == White)
         {
           my_HashCode ^= w_castle_status[(int)cs];
           my_HashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         }
         else
         {
           my_HashCode ^= b_castle_status[(int)cs];
           my_HashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         }
      }
   }
}

int Board::Between( const Square &sq1, const Square &sq2, 
                    Square *squares) const
{
   int offset = Bearing::Directions[sq1][sq2];
   if (offset == 0) return 0;
   Square sq = sq1;
   int n = 0;
   for (;;)
   {
      sq += offset;
      if (sq == sq2)
         break;
      ASSERT(OnBoard(sq));
      squares[n++] = sq;
   }
   return n;
}


BOOL Board::Clear( const Square &sq1, const Square &sq2 ) const
{
   int offset = Bearing::Directions[sq1][sq2];
   if (offset == 0) return 0;
   Square sq = sq1;
   for (;;)
   {
      sq += offset;
      if (sq == sq2)
         break;
      ASSERT(OnBoard(sq));
      if (my_Contents[sq] != EmptyPiece)
         return FALSE;
   }
   return TRUE;
}

/**
static inline void Xor( hash_t &h, const Square sq, const Piece piece )
{
   ASSERT(OnBoard(sq));
   h ^= hash_codes[sq][(int)piece];
}**/

#define Xor(h,sq,piece) h ^= hash_codes[sq][(int)piece]

void Board::DoMove( const Move &move )
{
   const ColorType oside = OppositeSide();
   const ColorType side = Side();
   my_CheckStatus = Unknown;
   move_count++;
   if (my_EnPassantSq[OppositeSide()] != InvalidSquare)
   {
       my_HashCode ^= ep_codes[my_EnPassantSq[OppositeSide()]];
       my_HashCode ^= ep_codes[0];
   }
   if (IsNull(move))
   {

      my_Side = oside;
      if (Side() == Black)
         my_HashCode |= (hash_t)1;
      else
         my_HashCode &= (hash_t)~1;
      *rep_list_head++ = my_HashCode;
      //ASSERT(rep_list_head-rep_list < 500);
      ASSERT(my_HashCode == Board_Hash::HashCode(*this));
      return;
   }

   ASSERT(!IsEmptyPiece(PieceMoved(move)));

   const Square start = StartSquare(move);
   const Square dest = DestSquare(move);
   const SpecialType moveType = TypeOfMove(move);
   if (side == White)
   {
      my_EnPassantSq[White] = InvalidSquare;

      if (moveType == KCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         Xor(my_HashCode, kp+3, whiteRook);
         Xor(my_HashCode, kp, whiteKing);
         Xor(my_HashCode, kp+1, whiteRook);
         Xor(my_HashCode, kp+2, whiteKing);
         my_HashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         my_HashCode ^= w_castle_status[(int)CastledKSide];

         const int newkp = kp + 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = whiteRook;
         my_Contents[newkp] = whiteKing;
         king_bits[White].clear(kp);
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         king_bits[White].set(newkp);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         Xor(my_HashCode, kp-4, whiteRook);
         Xor(my_HashCode, kp, whiteKing);
         Xor(my_HashCode, kp-1, whiteRook);
         Xor(my_HashCode, kp-2, whiteKing);
         my_HashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         my_HashCode ^= w_castle_status[(int)CastledQSide];

         const int newkp = kp - 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = Piece();
         my_Contents[newrooksq] = whiteRook;
         my_Contents[kp-2] = whiteKing;
         king_bits[White].clear(kp);
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         king_bits[White].set(newkp);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else // not castling
      {
         ASSERT(my_Contents[start] != EmptyPiece);
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured

         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(my_HashCode, start, whitePawn);
               Xor(my_HashCode, dest, whitePawn);
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               target = my_EnPassantSq[Black];
               my_Contents[dest] = whitePawn;
               pawn_bits[White].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(my_HashCode, start, whitePawn);
               Xor(my_HashCode, dest, MakeWhitePiece(PromoteTo(move)));
               Xor(m_pawn_hash, start, whitePawn);
               my_Contents[dest] = MakeWhitePiece(PromoteTo(move));
               my_Material[White].remove_pawn();
               my_Material[White].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[White].set(dest);
                  break;
               case Bishop:
                  bishop_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  break;
               case Rook:
                  rook_bits[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               case Queen:
                  queen_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(my_HashCode, start, whitePawn );
               Xor(my_HashCode, dest, whitePawn );
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               my_Contents[dest] = whitePawn;
               if (start - dest == 16) // 2-square pawn advance
               {
                  my_HashCode ^= ep_codes[0];
                  my_HashCode ^= ep_codes[dest];
                  my_EnPassantSq[White] = dest;
               }
               pawn_bits[White].set(dest);
               break;
            }
            pawn_bits[White].clear(start);
            break;
         case Knight:
            Xor(my_HashCode, start, whiteKnight);
            Xor(my_HashCode, dest, whiteKnight);
            my_Contents[dest] = whiteKnight;
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            Xor(my_HashCode, start, whiteBishop);
            Xor(my_HashCode, dest, whiteBishop);
            my_Contents[dest] = whiteBishop;
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            Xor(my_HashCode, start, whiteRook );
            Xor(my_HashCode, dest, whiteRook );
            my_Contents[dest] = whiteRook;
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            UpdateCastleStatus(side,start);
            break;
         case Queen:
            Xor(my_HashCode, start, whiteQueen);
            Xor(my_HashCode, dest, whiteQueen);
            my_Contents[dest] = whiteQueen;
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            Xor(my_HashCode, start, whiteKing );
            Xor(my_HashCode, dest, whiteKing );
            my_Contents[dest] = whiteKing;
            my_KingPos[White] = dest;
            if ((CastleStatus(White) != CastledQSide) &&
                (CastleStatus(White) != CastledKSide))
            {
               my_HashCode ^= w_castle_status[(int)CastleStatus(White)];
               my_HashCode ^= w_castle_status[(int)CantCastleEitherSide];
               my_CastleStatus[White] = CantCastleEitherSide;
            }
            king_bits[White].setClear(bits);
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece) 
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
            ASSERT(my_Contents[target] != EmptyPiece);
            occupied[Black].clear(target);
            Piece cap = MakeBlackPiece(Capture(move));
            Xor(my_HashCode, target, cap);
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[Black].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(Black,target);
               }
               my_Material[Black].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(Black,dest);
               rook_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               my_Material[Black].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[Black].clear(target);
               my_Material[Black].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Queen);
               break;
            case King:
               ASSERT(FALSE);
               king_bits[Black].clear(target);
               my_KingPos[Black] = InvalidSquare;
               my_CastleStatus[Black] = CantCastleEitherSide;
               my_Material[Black].remove_piece(King);
               break;
            default: 
               break;
            }
         }
      }
      setAll(White,dest);
      clearAll(White,start);
      my_EnPassantSq[Black] = InvalidSquare;

   }
   else // side == Black
   {
      my_EnPassantSq[Black] = InvalidSquare;
      if (moveType == KCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);

         // update the hash code
         Xor(my_HashCode, kp+3, blackRook);
         Xor(my_HashCode, kp, blackKing);
         Xor(my_HashCode, kp+1, blackRook);
         Xor(my_HashCode, kp+2, blackKing);
         my_HashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         my_HashCode ^= b_castle_status[(int)CastledKSide];

         const int newkp = kp + 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp+2] = blackKing;
         king_bits[Black].clear(kp);
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         king_bits[Black].set(newkp);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);

         // update the hash code
         Xor(my_HashCode, kp-4, blackRook);
         Xor(my_HashCode, kp, blackKing);
         Xor(my_HashCode, kp-1, blackRook);
         Xor(my_HashCode, kp-2, blackKing);
         my_HashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         my_HashCode ^= b_castle_status[(int)CastledQSide];

         const int newkp = kp - 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp-2] = blackKing;
         king_bits[Black].clear(kp);
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         king_bits[Black].set(newkp);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else // not castling
      {
  	     ASSERT(my_Contents[start] != EmptyPiece);
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured

         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(my_HashCode, start, blackPawn);
               Xor(my_HashCode, dest, blackPawn);
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               target = my_EnPassantSq[White];
               my_Contents[dest] = blackPawn;
               pawn_bits[Black].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(my_HashCode, start, blackPawn);
               Xor(my_HashCode, dest, MakeBlackPiece(PromoteTo(move)));
               Xor(m_pawn_hash, start, blackPawn);
               my_Contents[dest] = MakeBlackPiece(PromoteTo(move));
               my_Material[Black].remove_pawn();
               my_Material[Black].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[Black].set(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  break;
               case Rook:
                  rook_bits[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               case Queen:
                  queen_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(my_HashCode, start, blackPawn );
               Xor(my_HashCode, dest, blackPawn );
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               my_Contents[dest] = blackPawn;
               if (dest - start == 16) // 2-square pawn advance
               { 
                  my_HashCode ^= ep_codes[0];
                  my_HashCode ^= ep_codes[dest];
                  my_EnPassantSq[Black] = dest;
               }
               pawn_bits[Black].set(dest);
               break;
            }
            pawn_bits[Black].clear(start);
            break;
         case Knight:
            Xor(my_HashCode, start, blackKnight);
            Xor(my_HashCode, dest, blackKnight);
            my_Contents[dest] = blackKnight;
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            Xor(my_HashCode, start, blackBishop);
            Xor(my_HashCode, dest, blackBishop);
            my_Contents[dest] = blackBishop;
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            Xor(my_HashCode, start, blackRook );
            Xor(my_HashCode, dest, blackRook );
            my_Contents[dest] = blackRook;
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            UpdateCastleStatus(side,start);
            break;
         case Queen:
            Xor(my_HashCode, start, blackQueen);
            Xor(my_HashCode, dest, blackQueen);
            my_Contents[dest] = blackQueen;
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            Xor(my_HashCode, start, blackKing );
            Xor(my_HashCode, dest, blackKing );
            my_Contents[dest] = blackKing;
            my_KingPos[Black] = dest;
            if ((CastleStatus(Black) != CastledQSide) &&
                (CastleStatus(Black) != CastledKSide))
            {
               my_HashCode ^= b_castle_status[(int)CastleStatus(Black)];
               my_HashCode ^= b_castle_status[(int)CantCastleEitherSide];
               my_CastleStatus[Black] = CantCastleEitherSide;
            }
            king_bits[Black].setClear(bits);
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
            ASSERT(my_Contents[target] != EmptyPiece);
            occupied[White].clear(target);
            Piece cap = MakeWhitePiece(Capture(move));
            Xor(my_HashCode, target, cap);
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[White].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(White,target);
               }
               my_Material[White].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(White,dest);
               rook_bits[White].clear(target);
               rooksqueens[White].clear(target);
               my_Material[White].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[White].clear(target);
               my_Material[White].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[White].clear(target);
               rooksqueens[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Queen);
               break;
            case King:
               ASSERT(FALSE);
               king_bits[White].clear(target);
               my_KingPos[White] = InvalidSquare;
               my_CastleStatus[White] = CantCastleEitherSide;
               my_Material[White].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(Black,dest);
         clearAll(Black,start);
      }
      my_EnPassantSq[White] = InvalidSquare;
   }

   //UNCOND_ASSERT(rep_list_head-rep_list < 500);
   if (Side() == White)
      my_HashCode |= (hash_t)1;
   else
      my_HashCode &= (hash_t)~1;
   *rep_list_head++ = my_HashCode;
   ASSERT(pawn_hash() == Board_Hash::pawn_hash(*this));
   my_Side = OppositeSide();
   ASSERT(my_HashCode == Board_Hash::HashCode(*this));
   all_occupied = occupied[White];
   all_occupied.or(occupied[Black]);
#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy = *this;
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(king_bits[White] == copy.king_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(king_bits[Black] == copy.king_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);
#endif     
}

void Board::DoMoveQ( const Move &move )
{
   const ColorType oside = OppositeSide();
   const ColorType side = Side();
   my_CheckStatus = Unknown;
   move_count++;

   ASSERT(!IsNull(move));
   ASSERT(!IsEmptyPiece(PieceMoved(move)));

   const Square start = StartSquare(move);
   const Square dest = DestSquare(move);
   const SpecialType moveType = TypeOfMove(move);
   if (side == White)
   {
      my_EnPassantSq[White] = InvalidSquare;
      if (moveType == KCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         const int newkp = kp + 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = whiteRook;
         my_Contents[newkp] = whiteKing;
         king_bits[White].clear(kp);
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         king_bits[White].set(newkp);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;

         const Square kp = KingPos(White);
         const int newkp = kp - 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = Piece();
         my_Contents[newrooksq] = whiteRook;
         my_Contents[kp-2] = whiteKing;
         king_bits[White].clear(kp);
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         king_bits[White].set(newkp);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else // not castling
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               target = my_EnPassantSq[Black];
               my_Contents[dest] = whitePawn;
               pawn_bits[White].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(m_pawn_hash, start, whitePawn);
               my_Contents[dest] = MakeWhitePiece(PromoteTo(move));
               my_Material[White].remove_pawn();
               my_Material[White].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[White].set(dest);
                  break;
               case Bishop:
                  bishop_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  break;
               case Rook:
                  rook_bits[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               case Queen:
                  queen_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               my_Contents[dest] = whitePawn;
               if (start - dest == 16) // 2-square pawn advance
               {
                  my_EnPassantSq[White] = dest;
               }
               pawn_bits[White].set(dest);
               break;
            }
            pawn_bits[White].clear(start);
            break;
         case Knight:
            my_Contents[dest] = whiteKnight;
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            my_Contents[dest] = whiteBishop;
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            my_Contents[dest] = whiteRook;
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            UpdateCastleStatus(White,start);
            break;
         case Queen:
            my_Contents[dest] = whiteQueen;
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            my_Contents[dest] = whiteKing;
            my_KingPos[White] = dest;
            if ((CastleStatus(White) != CastledQSide) &&
                (CastleStatus(White) != CastledKSide))
            {
               my_CastleStatus[White] = CantCastleEitherSide;
            }
            king_bits[White].setClear(bits);
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
            occupied[Black].clear(target);
            Piece cap = MakeBlackPiece(Capture(move));
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[Black].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(Black,target);
               }
               my_Material[Black].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(Black,dest);
               rook_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               my_Material[Black].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[Black].clear(target);
               my_Material[Black].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Queen);
               break;
            case King:
               ASSERT(FALSE);
               king_bits[Black].clear(target);
               my_KingPos[Black] = InvalidSquare;
               my_CastleStatus[Black] = CantCastleEitherSide;
               my_Material[Black].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(White,dest);
         clearAll(White,start);
      }
      my_EnPassantSq[Black] = InvalidSquare;

   }
   else // side == Black
   {
      my_EnPassantSq[Black] = InvalidSquare;
      if (moveType == KCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);
         const int newkp = kp + 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp+2] = blackKing;
         king_bits[Black].clear(kp);
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         king_bits[Black].set(newkp);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);
         const int newkp = kp - 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp-2] = blackKing;
         king_bits[Black].clear(kp);
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         king_bits[Black].set(newkp);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else // not castling
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(m_pawn_hash, start, blackPawn);
               //Xor(m_pawn_hash, target, whitePawn);
               Xor(m_pawn_hash, dest, blackPawn);
               target = my_EnPassantSq[White];
               my_Contents[dest] = blackPawn;
               pawn_bits[Black].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(m_pawn_hash, start, blackPawn);
               my_Contents[dest] = MakeBlackPiece(PromoteTo(move));
               my_Material[Black].remove_pawn();
               my_Material[Black].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[Black].set(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  break;
               case Rook:
                  rook_bits[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               case Queen:
                  queen_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               my_Contents[dest] = blackPawn;
               if (dest - start == 16) // 2-square pawn advance
               {
                  my_EnPassantSq[Black] = dest;
               }
               pawn_bits[Black].set(dest);
               break;
            }
            pawn_bits[Black].clear(start);
            break;
         case Knight:
            my_Contents[dest] = blackKnight;
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            my_Contents[dest] = blackBishop;
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            my_Contents[dest] = blackRook;
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            UpdateCastleStatus(Black,start);
            break;
         case Queen:
            my_Contents[dest] = blackQueen;
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            my_Contents[dest] = blackKing;
            my_KingPos[Black] = dest;
            if ((CastleStatus(side) != CastledQSide) &&
                (CastleStatus(side) != CastledKSide))
               my_CastleStatus[Black] = CantCastleEitherSide;
            king_bits[Black].setClear(bits);
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
            occupied[White].clear(target);
            Piece cap = MakeWhitePiece(Capture(move));
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[White].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(White,target);
               }
               my_Material[White].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(White,dest);
               rook_bits[White].clear(target);
               rooksqueens[White].clear(target);
               my_Material[White].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[White].clear(target);
               my_Material[White].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[White].clear(target);
               rooksqueens[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Queen);
               break;
            case King:
               ASSERT(FALSE);
               king_bits[White].clear(target);
               my_KingPos[White] = InvalidSquare;
               my_CastleStatus[White] = CantCastleEitherSide;
               my_Material[White].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(Black,dest);
         clearAll(Black,start);
      }
      my_EnPassantSq[White] = InvalidSquare;
   }

   //UNCOND_ASSERT(rep_list_head-rep_list < 500);
   *rep_list_head++ = my_HashCode;
   ASSERT(pawn_hash() == Board_Hash::pawn_hash(*this));
   my_Side = OppositeSide();
   all_occupied = occupied[White];
   all_occupied.or(occupied[Black]);
#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy = *this;
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(king_bits[White] == copy.king_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(king_bits[Black] == copy.king_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);
#endif     
}

void Board::undo_castling(const Square &kp,
                          const Square &oldkingsq, const Square &newrooksq,
                          const Square &oldrooksq)
{
   my_Contents[kp] = EmptyPiece;
   my_Contents[oldrooksq] = MakePiece(Rook,my_Side);
   my_Contents[newrooksq] = EmptyPiece;
   my_Contents[oldkingsq] = MakePiece(King,my_Side);
   my_KingPos[my_Side] = oldkingsq;
   king_bits[my_Side].clear(kp);
   king_bits[my_Side].set(oldkingsq);
   rook_bits[my_Side].set(oldrooksq);
   rook_bits[my_Side].clear(newrooksq);
   rooksqueens[my_Side].set(oldrooksq);
   rooksqueens[my_Side].clear(newrooksq);

   setAll(my_Side,oldrooksq);
   setAll(my_Side,oldkingsq);
   clearAll(my_Side,kp);
   clearAll(my_Side,newrooksq);
}

void Board::UndoMove( const Move rmove, const Board_State &state )
{
   my_Side = OppositeColor(my_Side);
   if (!IsNull(rmove))
   {
      const ColorType opp_side = OppositeColor(my_Side);
      const SpecialType moveType = TypeOfMove(rmove);
      const Square start = StartSquare(rmove);
      const Square dest = DestSquare(rmove);
      if (moveType == KCastle)
      {
         Square kp = KingPos(my_Side);
         Square oldrooksq = kp+1;
         Square newrooksq = kp-1;
         Square oldkingsq = kp-2;
         undo_castling(kp,oldkingsq,newrooksq,oldrooksq);
      }
      else if (moveType == QCastle)
      {
         Square kp = KingPos(my_Side);
         Square oldrooksq = kp-2;
         Square newrooksq = kp+1;
         Square oldkingsq = kp+2;
         undo_castling(kp,oldkingsq,newrooksq,oldrooksq);
      }
      else if (my_Side == White)
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         // not castling
         Square target;
         if (moveType == EnPassant)
            target = DestSquare(rmove) + RankIncr;
         else
            target = DestSquare(rmove);
         // fix up start square:
         if (moveType == Promotion)
         {
            my_Contents[start] = whitePawn;
         }
         else
         {
            my_Contents[start] = MakeWhitePiece(PieceMoved(rmove));
         }
         setAll(White,start);
         switch (PieceMoved(rmove))
         {
         case Pawn:
            if (moveType == Promotion)
            {
               my_Material[White].add_pawn();
               my_Material[White].remove_piece(PromoteTo(rmove));
               switch (PromoteTo(rmove))
               {
               case Knight:
                  knight_bits[White].clear(dest);
                  break;
               case Bishop:
                  bishop_bits[White].clear(dest);
                  bishopsqueens[White].clear(dest);
                  break;
               case Rook:
                  rook_bits[White].clear(dest);
                  rooksqueens[White].clear(dest);
                  break;
               case Queen:
                  queen_bits[White].clear(dest);
                  bishopsqueens[White].clear(dest);
                  rooksqueens[White].clear(dest);
                  break;
               default:
                  break;
               }
            }
            else
            {
               pawn_bits[White].clear(dest);
            }
            pawn_bits[White].set(start);
            break;
         case Knight:
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            break;
         case Queen:
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            king_bits[White].setClear(bits);
            my_KingPos[White] = start;
            break;
         default:
            break;
         }
         // fix up dest square
         clearAll(White,dest);
         my_Contents[dest] = EmptyPiece;
         my_Contents[target] = MakePiece(Capture(rmove),Black);
         my_EnPassantSq[Black] = InvalidSquare;
         if (Capture(rmove) != EmptyPiece)
         {
            switch (Capture(rmove))
            {
            case Pawn:
               pawn_bits[Black].set(target);
               my_Material[Black].add_pawn();
               break;
            case Knight:
               knight_bits[Black].set(target);
               my_Material[Black].add_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].set(target);
               bishopsqueens[Black].set(target);
               my_Material[Black].add_piece(Bishop);
               break;
            case Rook:
               rook_bits[Black].set(target);
               rooksqueens[Black].set(target);
               my_Material[Black].add_piece(Rook);
               break;
            case Queen:
               queen_bits[Black].set(target);
               bishopsqueens[Black].set(target);
               rooksqueens[Black].set(target);
               my_Material[Black].add_piece(Queen);
               break;
            case King:
               king_bits[Black].set(target);
               my_KingPos[Black] = target;
               my_Material[Black].add_piece(King);
               break;
            default:
               break;
            }
            setAll(Black,target);
         }
      }
      else // my_Side == Black
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         // not castling
         Square target;
         if (moveType == EnPassant)
            target = dest - RankIncr;
         else
            target = dest;
         // fix up start square:
         if (moveType == Promotion)
         {
            my_Contents[start] = blackPawn;
         }
         else
         {
            my_Contents[start] = MakeBlackPiece(PieceMoved(rmove));
         }
         setAll(Black,start);
         switch (PieceMoved(rmove))
         {
         case Pawn:
            if (moveType == Promotion)
            {
               my_Material[Black].add_pawn();
               my_Material[Black].remove_piece(PromoteTo(rmove));
               switch (PromoteTo(rmove))
               {
               case Knight:
                  knight_bits[Black].clear(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].clear(dest);
                  bishopsqueens[Black].clear(dest);
                  break;
               case Rook:
                  rook_bits[Black].clear(dest);
                  rooksqueens[Black].clear(dest);
                  break;
               case Queen:
                  queen_bits[Black].clear(dest);
                  bishopsqueens[Black].clear(dest);
                  rooksqueens[Black].clear(dest);
                  break;
               default:
                  break;
               }
            }
            else
            {
               pawn_bits[Black].clear(dest);
            }
            pawn_bits[Black].set(start);
            break;
         case Knight:
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            break;
         case Queen:
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            king_bits[Black].setClear(bits);
            my_KingPos[Black] = start;
            break;
         default:
            break;
         }
         // fix up dest square
         clearAll(Black,dest);
         my_Contents[dest] = EmptyPiece;
         my_Contents[target] = MakePiece(Capture(rmove),White);
         my_EnPassantSq[White] = InvalidSquare;
         if (Capture(rmove) != EmptyPiece)
         {
            switch (Capture(rmove))
            {
            case Pawn:
               pawn_bits[White].set(target);
               my_Material[White].add_pawn();
               break;
            case Knight:
               knight_bits[White].set(target);
               my_Material[White].add_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].set(target);
               bishopsqueens[White].set(target);
               my_Material[White].add_piece(Bishop);
               break;
            case Rook:
               rook_bits[White].set(target);
               rooksqueens[White].set(target);
               my_Material[White].add_piece(Rook);
               break;
            case Queen:
               queen_bits[White].set(target);
               bishopsqueens[White].set(target);
               rooksqueens[White].set(target);
               my_Material[White].add_piece(Queen);
               break;
            case King:
               king_bits[White].set(target);
               my_KingPos[White] = target;
               my_Material[White].add_piece(King);
               break;
            default:
               break;
            }
            setAll(White,target);
         }
      }
   }
   my_EnPassantSq[Black] = state.old_enpassantsq[Black];
   my_EnPassantSq[White] = state.old_enpassantsq[White];
   my_CastleStatus[Black] = state.old_castlestatus[Black];
   my_CastleStatus[White] = state.old_castlestatus[White];
   my_CheckStatus = state.old_checkStatus;
   my_HashCode = state.old_hashcode;
   m_pawn_hash = state.old_pawn_hashcode;
   move_count = state.old_movecount;
   --rep_list_head;
//    UNCOND_ASSERT(rep_list_head >= rep_list);
   all_occupied = occupied[White];
   all_occupied.or(occupied[Black]);
#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy = *this;
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(king_bits[White] == copy.king_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(king_bits[Black] == copy.king_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);
#endif     
}

BOOL Board::any_attacks(const Square sq, const ColorType side) const
{
   if (sq == InvalidSquare)
      return FALSE;

   if (side == White)
   {
      if (and(Bearing::pawn_attacks[sq][White],pawn_bits[White]))
         return TRUE;

      if (and(Bearing::knight_attacks[sq],knight_bits[White]))
         return TRUE;

      if (and(Bearing::king_attacks[sq],king_bits[White]))
         return TRUE;
   }
   else
   {
      if (and(Bearing::pawn_attacks[sq][Black],pawn_bits[Black]))
         return TRUE;

      if (and(Bearing::knight_attacks[sq],knight_bits[Black]))
         return TRUE;

      if (and(Bearing::king_attacks[sq],king_bits[Black]))
         return TRUE;
   }

   // Find attacks on the rank
   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (!Bitmap::and(rooksqueens[side],Bearing::rank_mask[r]).is_clear())
   {
      // See if any rooks or queens are there
      if (!Bitmap::and(RankAttacks(*this,sq),rooksqueens[side]).is_clear())
         return TRUE;
   }

   if (!Bitmap::and(rooksqueens[side],Bearing::file_mask[f]).is_clear())
   {
      if (!Bitmap::and(FileAttacks(*this,sq),rooksqueens[side]).is_clear())
         return TRUE;
   }

   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a1_mask[sq]).is_clear())
   {
      // Now find attacks on the a1-h8 diagonal
      if (!Bitmap::and(DiagAtcksA1(*this,sq),bishopsqueens[side]).is_clear())
         return TRUE;
   }

   // Now find attacks on the a8-h1 diagonal
   // See if any bishops or queens are there
   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a8_mask[sq]).is_clear())
   {
      if (!Bitmap::and(DiagAtcksA8(*this,sq),bishopsqueens[side]).is_clear())
         return TRUE;
   }
   return FALSE;
}

BOOL Board::any_attacks(const Square sq, const ColorType side,
                        Bitmap &source) const
{
   if (sq == InvalidSquare)
      return FALSE;

   source = Bearing::pawn_attacks[sq][side];
   source.and(pawn_bits[side]);
   if (!source.is_clear())
      return TRUE;

   source = Bearing::knight_attacks[sq];
   source.and(knight_bits[side]);
   if (!source.is_clear())
      return TRUE;

   source = Bearing::king_attacks[sq];
   source.and(king_bits[side]);
   if (!source.is_clear())
      return TRUE;

   // Find attacks on the rank
   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (!Bitmap::and(rooksqueens[side],Bearing::rank_mask[r]).is_clear())
   {
      source = Bitmap::and(RankAttacks(*this,sq),rooksqueens[side]);
      if (!source.is_clear())
         return TRUE;
   }

   Bitmap result;
   if (!Bitmap::and(rooksqueens[side],Bearing::file_mask[f]).is_clear())
   {
      if (!Bitmap::and(FileAttacks(*this,sq),rooksqueens[side]).is_clear())
         return TRUE;
   }

   // Now find attacks on the a1-h8 diagonal
   // find the pieces nearest to sq on the diagonal, then unrotate the bits
   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a1_mask[sq]).is_clear())
   {
      source = DiagAtcksA1(*this,sq);
      // See if any bishops or queens are there
      if (!source.is_clear())
         return TRUE;
   }

   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a8_mask[sq]).is_clear())
   {
      // Now find attacks on the a8-h1 diagonal
      source = DiagAtcksA8(*this,sq);
      if (!source.is_clear())
         return TRUE;
   }
   return FALSE;
}

Bitmap Board::calc_attacks(const Square sq, const ColorType side) const
{
   Bitmap retval;

   Bitmap patcks(Bearing::pawn_attacks[sq][side]);
   patcks.and(pawn_bits[side]);
   retval.or(patcks);

   Bitmap natcks(Bearing::knight_attacks[sq]);
   natcks.and(knight_bits[side]);
   retval.or(natcks);

   Bitmap katcks(Bearing::king_attacks[sq]);
   katcks.and(king_bits[side]);
   retval.or(katcks);

   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (!Bitmap::and(rooksqueens[side],Bearing::rank_mask[r]).is_clear())
   {
      retval.data |= Bitmap::and(RankAttacks(*this,sq),rooksqueens[side]).data;
   }

   if (!Bitmap::and(rooksqueens[side],Bearing::file_mask[f]).is_clear())
   {
      // To find attacks on the file, use the 90-degree rotated bitmaps.
      // Same procedure as for ranks, above.
      // shift the occupied bits so that the file that sq occupies are
      // in bits 0-7.
      // find the pieces nearest to sq on the file, then
      // "unrotate" the result and shift it back up.
      // See if any rooks or queens are there
      retval.data |= Bitmap::and(FileAttacks(*this,sq),rooksqueens[side]).data;
   }

   // Now find attacks on the a1-h8 diagonal
   // find the pieces nearest to sq on the diagonal, then unrotate the bits
   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a1_mask[sq]).is_clear())
   {
      retval.data |= Bitmap::and(DiagAtcksA1(*this,sq),
                            bishopsqueens[side]).data;
   }
   if (!Bitmap::and(bishopsqueens[side],Bearing::diag_a8_mask[sq]).is_clear())
   {
      // Now find attacks on the a8-h1 diagonal
      retval.data |= Bitmap::and(DiagAtcksA8(*this,sq),
                            bishopsqueens[side]).data;
   }
   return retval;
}

Bitmap Board::min_attacker(const Bitmap &atcks, const ColorType side)
const
{
   Bitmap retval;
   if (side == White)
   {
      retval = Bitmap::and(atcks,pawn_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,knight_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,bishop_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,rook_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,queen_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,king_bits[White]);
   }
   else
   {
      retval = Bitmap::and(atcks,pawn_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,knight_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,bishop_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,rook_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,queen_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::and(atcks,king_bits[Black]);
   }    
   return retval;
}

Bitmap Board::pawn_attacks( const Square &sq, const ColorType side) const
{
   Bitmap patcks(Bearing::pawn_attacks[sq][side]);
   patcks.and(pawn_bits[side]);
   return patcks;
}

Bitmap Board::all_pawn_attacks( const ColorType side) const
{
   if (side == White)
   {
      Bitmap pawns1(pawn_bits[White]);
      Bitmap pawns2(pawns1);
      pawns1.shr(7);
      pawns1.andNot(Bitmap(0x0101010101010101));
      pawns2.shr(9);
      pawns2.andNot(Bitmap(0x8080808080808080));
      return Bitmap::or(pawns1,pawns2);
   }
   else
   {
      Bitmap pawns1(pawn_bits[Black]);
      Bitmap pawns2(pawns1);
      pawns1.shl(7);
      pawns1.andNot(Bitmap(0x8080808080808080));
      pawns2.shl(9);
      pawns2.andNot(Bitmap(0x0101010101010101));
      return Bitmap::or(pawns1,pawns2);
   }
}

int Board::rook_mobility(const Square &sq) const
{
   // We consider only horizontal mobility (for now)
   int mobl;
   Bitmap occ(all_occupied);
   // Shift so that the relevant bits are in bits 0-7 of the bitmap.
   int r = ((int)sq)/8;
   mobl = Bearing::mobility[
       Bearing::whichAttack[(occ.rank_bits(r)) & 0xff][File(sq)-1]];
   return mobl;
}

Board::CheckStatusType Board::CheckStatus() const
{
   if (my_CheckStatus != Unknown)
   {
      return my_CheckStatus;
   }
   else if (any_attacks(KingPos(Side()),OppositeSide()))
   {
      // This is a const function, but we cache its result
      Board &b = (Board&)*this;
      b.my_CheckStatus = InCheck;
   }
   else
   {
      // This is a const function, but we cache its result
      Board &b = (Board&)*this;
      b.my_CheckStatus = NotInCheck;
   }
   return my_CheckStatus;
}

static void set_bad( istream &i )
{
   i.clear( ios::badbit | i.rdstate() );
}

int Board::rep_count() const
{
   int entries = move_count-2;
   hash_t to_match = HashCode();
   int count = 0;
   for (hash_t *rep_list=rep_list_head-3;
       entries>=0;
       rep_list-=2,entries-=2)
   {
      if (*rep_list == to_match)
      {
         count++;
         if (count >= 2)
         {
            return count;
         }
      }
   }
   return count;
}

istream & operator >> (istream &i, Board &board)
{
   // read in a board position in Forsythe-Edwards (FEN) notation.
   static char buf[128];

   char *bp = buf;
   int c;
   int fields = 0; int count = 0;
   while (i.good() && fields < 4 && (c = i.get()) != '\n' && 
          ++count < 128)
   {
      *bp++ = c;
      if (isspace(c))
         fields++;
   }
   *bp = '\0';
   if (!i)
      return i;
   if (!BoardIO::read_fen(board, buf))
      set_bad(i);
   return i;
}

ostream & operator << (ostream &o, const Board &board)
{
   char buf[128];
   memset(buf,'\0',128);
   BoardIO::write_fen(board,buf,TRUE);
   o << buf;
   return o;
}







