// Copyright 1993-1999 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "bookread.h"
#include "constant.h"
#include "bookdefs.h"
#include "movegen.h"
#include "moveord.h"
#include "globals.h"
#include "util.h"
#include "config.h"
#include <string.h>
#include <iostream.h> // for debugging
#ifdef _CONSOLE
#include <windows.h>
#else
#include <afxwin.h>
#endif

Book_Reader::Book_Reader(const char* fileName, const char* mappingName,
 bool create)
: book_file(NULL),
pBook(NULL),
pPage(NULL),
hFileMapping(NULL),
open(FALSE),
current_page(-1)
{                 

   static char book_path[MAX_PATH];
   strcpy(book_path,programPath);
   {
      // find the last part of the program pathname
      char *p = strrchr(book_path,'\\');
      if (p)
      {
         // replace it with the book file name
         strcpy(p+1,fileName);
      }
      else
         strcpy(book_path,fileName);
      book_file = BookUtil::openFile(book_path,false);
      if (book_file == NULL)
      {
        char msg[255];
        sprintf(msg,"%s %s","Warning: could not open book file ",fileName);
#ifdef _CONSOLE
         ::MessageBox(NULL,msg,"",MB_OK | MB_ICONINFORMATION);
#else
         AfxMessageBox(msg,MB_OK | MB_ICONINFORMATION);
#endif
      }
   }
   if (book_file != NULL)
   {
      hFileMapping = BookUtil::createMap(book_file,false);
      if (hFileMapping == NULL)
      {
#ifdef _CONSOLE
         ::MessageBox(NULL,"Error: could not create file mapping for opening book","",MB_OK | MB_ICONEXCLAMATION);
#else
         AfxMessageBox("Error: could not create file mapping for opening book",MB_OK | MB_ICONEXCLAMATION);
#endif
      }
   }
   if (hFileMapping)
   {
      pBook = BookUtil::map(hFileMapping,0,4096,false);
      if (pBook == NULL)
      {
#ifdef _CONSOLE
         ::MessageBox(NULL,"Error: file mapping failed for opening book","",MB_OK | MB_ICONEXCLAMATION);
#else
         AfxMessageBox("Error: file mapping failed for opening book",MB_OK | MB_ICONEXCLAMATION);
#endif
      }
   }

   open = pBook != NULL;
   if (!open)
      return;

   //memcpy(&hdr,(char*)pBook,Header_Size);
   char *p = (char*)pBook;
   hdr.version = (byte)(*p);
   hdr.num_pages = (byte)p[1];
   memcpy(&hdr.page_capacity,p+2,4);
   memcpy(&hdr.hash_table_size,p+6,4);

   if (hdr.version != (byte)Book_Version)
   {
      char msg[128];
      sprintf(msg,"Expected book version %d, got %d instead",
              (int)Book_Version,(int)hdr.version);
      MessageBox(NULL,msg,"Warning",MB_ICONEXCLAMATION);
      BookUtil::unmap(pBook);
      BookUtil::closeMap(hFileMapping);
      BookUtil::closeFile(book_file);
      open = FALSE;
      return;
   }
   BookUtil::unmap(pBook);
   pBook = NULL;
}

Book_Reader::~Book_Reader()
{
   if (pBook)
      BookUtil::unmap(pBook);
   if (hFileMapping)
      BookUtil::closeMap(hFileMapping);
   if (book_file)
      BookUtil::closeFile(book_file);
}

const void Book_Reader::head( const Board & b, enum styles style,
                              Book_Location &loc)
{
   loc.page = (int)(b.HashCode() % hdr.num_pages);
   loc.index = INVALID;
   fetch_page(loc.page);
   unsigned probe = (unsigned)((b.HashCode()>>32) % hdr.hash_table_size);
   if (style != Balanced)
   {
      // Make the low-level bits of the bucket index match the
      // style
      probe &= ~(MAX_STYLES-1);
      probe |= (int)style;
   }

   // Copy the index from the book
   memcpy(&loc.index,pPage+Header_Size+sizeof(uint16)*probe,sizeof(uint16));
}

void Book_Reader::fetch( const Book_Location &loc, Book_Info &book_info )
{
   ASSERT(pPage);

   byte *entry = pPage + Header_Size + hdr.hash_table_size*2 + ((int)loc.index)*Entry_Size;
   hash_t hc;
   byte freq, rec, indx;
   // We assume here that the compiler doesn't reorder structures

   memcpy(&hc,entry,sizeof(hash_t));
   freq = entry[8];
   indx = entry[9];
   byte flags = entry[10];
   rec = flags & 0x7;
   book_info.init(hc,freq,flags,indx);
   book_info.winloss = entry[11];
   book_info.set_location(loc);
   memcpy(&(book_info.learn_score),entry+12,4);
}

void Book_Reader::update(const Book_Location &loc, float learn_factor)
{
#ifdef LEARNING
   fetch_page(loc.page);
   byte *entry = pPage + Header_Size + hdr.hash_table_size*2 + ((int)loc.index)*Entry_Size;
   // get the existing learn factor
   float old_factor;
   memcpy(&old_factor,entry+12,4);
   learn_factor += old_factor;
   // update the learn factor
   memcpy(entry+12,&learn_factor,4);
#endif
}

static int get_weight(const Book_Info &be,int total_freq,const ColorType side)
{
   int rec = be.get_recommend();
   if (rec != 50)
      return rec; // weight from the old book
   int w = be.get_winloss();
   int freqWeight = (int)((100L*be.get_frequency())/total_freq);
   // Slightly inflate the frequency of rare moves
   if (freqWeight < 10)
      freqWeight = freqWeight + (10-freqWeight)/4;
   TRACE("weight = %d\n",
         freqWeight);
   int winWeight = (w+100)/2;
   // Favor more frequent moves and moves that win
   int base = (freqWeight*winWeight)/200;
   if (base == 0) return 0;
   // Factor in learning
#ifdef LEARNING
   int score = (int)be.learn_score;
   if (score >= 0)
      return Util::Min(100,(base*(10+score))/10);
   else
      return Util::Max(0,(base*(10+score))/10);
#else
   return base;
#endif
}

Move Book_Reader::pick( const Board &b, const Book_Location &loc,
                        Book_Info &info )
{
   unsigned total_weight = 0;
   Book_Location tmp = loc;
   int total_freq = 0;
   int count = 0;
   // get the total move frequency
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp, be);
      if (be.hash_code() == b.HashCode() && be.get_recommend() != 0)
      {
         total_freq += be.get_frequency();
         ++count;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   tmp = loc;
   // Determine the total weights of moves for this position.
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp, be);
      if (b.HashCode() == be.hash_code())
      {
         int w  = get_weight(be,total_freq,b.Side());
         total_weight += w;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   
   // If total_weight is 0, no moves have non-zero weights.
   // Return without any move selected.
   if (total_weight == 0)
      return NullMove;
   unsigned weight = 0;
   tmp = loc;
   const unsigned nRand = (rand()*(unsigned long)total_weight)/RAND_MAX;
   // Randomly pick from the available moves.  Prefer moves
   // with high weights.
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      int w;
      if (b.HashCode() == be.hash_code() && (w = get_weight(be,total_freq,b.Side())) > 0)
      {
         weight += w;
         if (nRand < weight)
         {
            // We have selected a move. The book contains
            // only the move index, not the move itself.
            // We must call the move generator to obtain the
            // move.
            Move moves[Move_Generator::MaxMoves];
            Move_Generator mg( b, 0, NullMove);
            int n = mg.Generate_Moves(moves,TRUE /*repeatable*/); 
            ASSERT(be.move_index < n);
            be.set_move(moves[be.move_index]);
            info = be;
            info.set_total_moves(count);
            info.set_location(tmp);
            return be.get_move();
         }
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   // should never get here
   ASSERT(FALSE);
   return NullMove;
}

int Book_Reader::book_moves(const Board &b, Move *moves, const unsigned limit)
{
   Book_Location tmp;
   head(b,Balanced,tmp);
   Book_Entry target(b.HashCode(),0,0,0);
   int num_moves = 0;
   int total_freq = 0;
   // get the total move frequency
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      if (target.is_equal(be) && be.get_recommend())
      {
         num_moves++;
         total_freq += be.get_frequency();
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   if (num_moves == 0)
      return 0;
   Move *tmp_moves = new Move[Move_Generator::MaxMoves];
   Move_Generator mg( b, 0, NullMove);
   int n = mg.Generate_Moves(tmp_moves,TRUE /* repeatable */); 
   int * scores = new int[num_moves];
   Move * tmp_moves2 = new Move[num_moves];
   Book_Location loc;
   head(b,Balanced,loc);
   num_moves = 0;
   tmp = loc;
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      if (target.is_equal(be) && be.get_recommend())
      {
         ASSERT(be.move_index < n);
         tmp_moves2[num_moves] = tmp_moves[be.move_index];
         scores[num_moves] = get_weight(be,total_freq,b.Side());
         ++num_moves;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   Move_Ordering::sort_moves(tmp_moves2,scores,num_moves);
   unsigned ret_val = Util::Min(num_moves,limit);
   for (unsigned i = 0; i < ret_val; i++)
      moves[i] = tmp_moves2[i];
   delete [] scores;
   delete [] tmp_moves;
   delete [] tmp_moves2;
   return ret_val;
}

void Book_Reader::fetch_page(int page)
{
   if (current_page != page)
   {
      if (pBook)
      {
         BookUtil::unmap(pBook);
      }
      pBook = BookUtil::map(hFileMapping,
                      page*hdr.page_capacity,hdr.page_capacity,false);
      if (!pBook)
      {
         MessageBox( NULL, "File mapping failed for opening book", "Warning", MB_ICONEXCLAMATION);
         pPage = NULL;
         return;
      }
      pPage = (byte*)pBook;
   }
   current_page = page;
}
