// Copyright 1994, 1995, 1999 by Jon Dart.

// This module handles reading and writing board and
// game info.

#include "stdafx.h"
#include "boardio.h"
#include "chessio.h" 
#include "legal.h"
#include "strpair.h"
#include "bearing.h"
#include "notation.h"
#ifdef _WINDOWS
#include "resource.h"
#endif
#include "log.h"
#include "search.h"
#include "globals.h"
#include "options.h"
#include <ctype.h>
#include <string.h>
#include <strstrea.h>
#include <stdio.h>
#include <time.h>

extern Options *global_options; 

const int MAX_TAG = 255;
const int MAX_VAL = 255;
const int PGN_MARGIN = 70; // max length of PGN line for output

static int skip_space(ifstream &game_file)
{
    int c = EOF;
    while (!game_file.eof())
    {
       c = game_file.get();
       if (!isspace(c) && (c!='\n'))
       {
           break;
       }
    }
    return c;
}

static BOOL get_header(CPtrArray &hdrs, LPCSTR key, CString &val)
{
      const int n = hdrs.GetSize();
      for (int i=0;i<n;i++)
      {
          StringPair *p = (StringPair*)hdrs[i];
          if (strcmp(p->tag,key)==0)
          {
               val = p->val;
               return TRUE;
          }
      }
      return FALSE;
}

static void add_header(CPtrArray &hdrs, LPCSTR key, LPCSTR val)
{
   ASSERT(key);
   ASSERT(val);
   hdrs.Add(new StringPair(key,val));
}

void ChessIO::get_game_description(CPtrArray &hdrs, CString &descr)
{
      CString tmp;
      get_header(hdrs, "White", tmp);
      int comma = tmp.Find(',');
      if (comma != -1)
      {
          tmp = tmp.Mid(0,comma);
      }
      descr = tmp;
      descr += "-";
      get_header(hdrs, "Black", tmp);
      comma = tmp.Find(',');
      if (comma != -1)
      {
          tmp = tmp.Mid(0,comma);
      }
      descr += tmp;
      get_header(hdrs, "Site", tmp);
      descr += ", ";
      descr += tmp;
      get_header(hdrs, "Round", tmp);
      if (tmp.GetLength() >0 && tmp[0] != '?')
      {
          descr += '(';
          descr += tmp;
          descr += ')';
      }
      descr += ' ';
      get_header(hdrs, "Date", tmp);
      descr += tmp.Mid(0,4);
      get_header(hdrs, "Result", tmp);
      if (tmp != "*") 
      {
         descr += " ";
         descr += tmp;
      }
}

void ChessIO::collect_headers(ifstream &game_file, CPtrArray *&hdrs, long &first)
{
        hdrs = new CPtrArray();
        first = -1L;
        int c;
        BOOL firstTag = TRUE;
        while (!game_file.eof())
        {
            char tag[MAX_TAG+1];
            char val[MAX_VAL+1];
            c = skip_space(game_file);
            if (c!='[')
            {
                game_file.putback(c);
                break;
            }
            else
            {
                if (first == -1)
                    first = game_file.tellg()-1;
                int t = 0;
                c = skip_space(game_file);
                while (!game_file.eof())
                {
                    if (!isspace(c) && c != '[' &&
                        c != '"' && t < MAX_TAG)
                        tag[t++] = c;
                    else
                        break;
                    c = game_file.get();
                } 
                tag[t] = '\0';
                if (firstTag)
                {
                    if (strcmp(tag,"vent") == 0)
                    {
                       // It appears that there is a bug in iostream.
                       // Calling tellg (above) sometimes causes
                       // a missing char when the next read occurs.
                       // So we get a partial tag.  This is a workaround.
                       strcpy(tag,"Event");
                    }
                    else if (strcmp(tag,"Event") != 0)
                    {
                       if (c == '[') game_file.putback(c);
                       continue;
                    }
                }
                firstTag = FALSE;
                if (isspace(c))
                    c = skip_space(game_file);
                if (c=='"')
                {
                    int v = 0;
                    while (!game_file.eof())
                    {
                        c = game_file.get();
                        if (c != '"' && v < MAX_VAL)
                            val[v++] = c;
                        else
                            break;
                    }
                    val[v] = '\0';
                }
                //else
                //    cerr << "bad tag" << endl;
                StringPair *new_hdr = new StringPair(tag,val);
                hdrs->Add(new_hdr);
                while (!game_file.eof() && c != ']')
                    c = game_file.get();
            }
        }
}

BOOL ChessIO::scan_pgn(ifstream &game_file, CStringList &contents)
{
    Board board;
    ColorType side = White;
    int c;

    while (game_file.good() && !game_file.eof())
    {
        // Collect the header:
        long first;
        CPtrArray *hdrs;
        CString eventStr;
        collect_headers(game_file, hdrs, first);
        if (get_header(*hdrs,"Event",eventStr))
        {
           // We have the headers, munge them into a single-line game
           // description
           CString descr;
           get_game_description(*hdrs, descr);
           char tmp[20];
           sprintf(tmp,"\t%ld",first);
           descr += tmp;
           contents.AddTail(descr);
        }
        // free the header info we allocated
        const int n = hdrs->GetSize();
        for (int i =0; i<n; i++)
        {
           StringPair *p = (StringPair*)hdrs->GetAt(i);
           delete p;
        }
        hdrs->RemoveAll();
        delete hdrs;
        
        while (game_file.good() && !game_file.eof())
        {
           if ((c = game_file.get()) == '[')
           {
               game_file.putback(c);
               break;
           }
        }
    } 
    return TRUE;
}

ChessIO::Token ChessIO::get_next_token(ifstream &game_file, char *buf, int limit)
{
    Token tok = Unknown;
    int count = 0;
    int c = skip_space(game_file);
    if (c == EOF)
       return Eof;
    else if (c=='{')
    {
        *buf++ = c;
        ++count;
        while (game_file.good() && c != '}')
        {
            c = game_file.get();
            if (count < limit-1)
            {
               *buf++ = c;
               ++count;
            }
        }
        if (c=='}' && count < limit-1)
        {
            *buf++ = c;
        }
        *buf = '\0';
        return Comment;
        //c = game_file.get();
    }
    if (isdigit(c))
    {
        int nextc = game_file.get();
        if (c == '0')
        {
           // peek at next char.
           if (nextc == '-')
           {
               // some so-called PGN files have 0-0 or 0-0-0 for
               // castling.  To handle this, we need to peek ahead
               // one more character.
               int nextc2 = game_file.peek();
               if (toupper(nextc2) == 'O' || nextc2 == '0')
               {            
                  // castling, we presume
                  buf[count++] = c;
                  c = nextc;
                  while (game_file.good() && (c == '-' ||
                    c == '0' || toupper(c) == 'O' || c == '+'))
                  {
                     if (count < limit-1)
                        buf[count++] = c;
                     c = game_file.get();
                  }
                  buf[count] = '\0';
                  game_file.putback(c);
                  return GameMove;
               }
           }
        }
        if (nextc == '-' || nextc == '/') // assume result
        {
            buf[count++] = c;
            c = nextc;
            while (!game_file.eof() && game_file.good() && !isspace(c))
            {
                if (count < limit-1)
                    buf[count++] = c;
                c = game_file.get();
            }
            tok = Result;
        }
        else
        {
            // Assume we have a move number.
            buf[count++] = c;
            c = nextc;
            while (game_file.good() && (isdigit(c) || c == '.'))
            {
               if (count < limit-1)
                  buf[count++] = c;
               c = game_file.get();
            }
            game_file.putback(c);
            return Number;
       }
       buf[count] = '\0';
   }           
   else if (isalpha(c))
   {
       while (game_file.good() && (isalnum(c) 
              || c == '-' || c == '=' || (c == '+')))
       {
           if (count < limit-1)
               buf[count++] = c;
           c = game_file.get();
       }
       buf[count] = '\0';
       game_file.putback(c);
       return GameMove;
   }
   else if (c == '#') // "Checkmate"
   {
       *buf = c; *(buf+1) = '\0';
       tok = Ignore;
   }
   else if (c == '*')
   {
       *buf = c; *(buf+1) = '\0';
       tok = Result;
   }
   else
   {
       *buf = c; *(buf+1) = '\0';
       tok = Unknown;
   }
   return tok;
} 

#ifndef _CONSOLE
BOOL ChessIO::load_pgn(Board &board,ifstream &game_file, Log &log, DWORD offset,
CPtrArray *&hdrs)
{
    board.Reset();
    ColorType side = White;
    game_file.seekg(offset); 
    // seekg does not clear the eof bit:
    game_file.clear();

    // Collect the header:
    long first;
    collect_headers(game_file, hdrs, first);

    // See if the file contained a "FEN" tag.
    CString FENString;
    if (get_header(*hdrs,"FEN",FENString))
    {
       char *buf = FENString.GetBuffer(256);

       istrstream FENstream(buf,256);
       if (!load_fen(FENstream,board))
       {
          // some error in the FEN
#ifdef _WINDOWS
          CString msg;
          msg.LoadString(IDS_INVALID_FEN);
          ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
#endif
          FENString.ReleaseBuffer();
          return FALSE;
       }
       FENString.ReleaseBuffer();
    }

    // read the moves
    Board tmp_board(board);
    for (;;)
    {
        static char buf[256];
        static char num[20];
        Token tok = get_next_token(game_file,buf,256);
        if (tok == Eof)
             break;
        if (tok == Number)
        {
            strncpy(num,buf,20);
        }
        else if (tok == GameMove)
        {
            // parse the move
            Move m = Notation::Value(tmp_board,side,buf);
            if (IsNull(m) ||
                !legal_move(tmp_board,StartSquare(m),
                            DestSquare(m)))
            {
#ifdef _WINDOWS
                CString msg;
                AfxFormatString1(msg,IDS_ILLEGAL_MOVE,buf);
                ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
                log.clear();
#endif

                break;
            }
            else
            {
                //MessageBox(NULL,m.Image(),"",MB_OK);
#ifdef _WINDOWS
                log.add_move(tmp_board,m,buf,NULL,
                             NULL,FALSE);
#endif
                tmp_board.DoMove(m);
            }
            side = OppositeColor(side);
         }
         else if (tok == Unknown)
         {
#ifdef _WINDOWS
             CString msg;
             AfxFormatString1(msg,IDS_UNREC_TEXT,buf);
             ::MessageBox(NULL,(LPCSTR)msg,"",MB_OK);
#endif
         }
         else if (tok == Comment)
         {
             // ignored for now
         }
         else if (tok == Result)
         {
            if (*buf != '*')
               log.setResult(buf);
            break;
         }            
    } 
    return TRUE;
}
#endif

BOOL ChessIO::load_fen(istream &ifs, Board &board)
{                  
    ifs >> board;
    board.rep_list_head = board.rep_list;
    board.move_count = 0;
    return ifs.good();
}

BOOL ChessIO::store_pgn(ostream &ofile,Log &log,
                        const ColorType computer_side,
                        CPtrArray &headers)
{
    // Write standard PGN header.
    CString result = log.getResult();
    if (result.GetLength() == 0)
       result = "*";
    CString val;
    CPtrArray newHeaders;
    if (!get_header(headers, "Event", val))
       add_header(newHeaders,"Event","?");
    else
       add_header(newHeaders,"Event",val);
    if (!get_header(headers, "Site", val))
       add_header(newHeaders,"Site","?");
    else
       add_header(newHeaders,"Site",val);
    if (!get_header(headers, "Date", val))
    {
       char string[15];
       time_t tm = time(NULL);
       struct tm *t = localtime(&tm);
       sprintf(string,"%4d.%02d.%02d",t->tm_year+1900,t->tm_mon+1,
                        t->tm_mday);
       add_header(newHeaders,"Date",string);
    }
    else
       add_header(newHeaders,"Date",val);
    if (!get_header(headers, "Round", val))
       add_header(newHeaders,"Round","?");
    else
       add_header(newHeaders,"Round",val);
 
    char name[30];
    sprintf(name,"%s %s","Arasan",Arasan_Version);
    if (computer_side == White)
    {
       if (!get_header(headers, "White",val))
       {
         add_header(newHeaders, "White", name);
       }
       else
         add_header(newHeaders,"White",val);
       if (!get_header(headers, "Black",val))
         add_header(newHeaders,"Black","?");
       else
         add_header(newHeaders,"Black",val);
    } 
    else
    {
       if (!get_header(headers, "White",val))
         add_header(newHeaders,"White","?");
       else
         add_header(newHeaders,"White",val);
       if (!get_header( headers, "Black",val))
         add_header(newHeaders, "Black", name);
       else
         add_header(newHeaders,"Black",val);
    }
    // "Result" may contain a comment. Don't put this in the
    // PGN header.
    CString shortResult = result;
    int space = result.Find(' ');
    if (space >0)
       shortResult = result.Mid(0,space);
    if (!get_header(headers,"Result",val))
       add_header(newHeaders,"Result",shortResult);
    else
       add_header(newHeaders,"Result",val);

    // We have now written all the mandatory headers.
    // Add any more headers that were passed into us.

    int n = headers.GetSize();
    for (int i=0;i<n;i++)
    {
        StringPair *p = (StringPair*)headers[i];
        if (strcmp(p->tag,"Event")!=0 &&
            strcmp(p->tag,"Site")!=0 &&
            strcmp(p->tag,"Date")!=0 &&
            strcmp(p->tag,"Round")!=0 &&
            strcmp(p->tag,"White")!=0 &&
            strcmp(p->tag,"Black")!=0 &&
            strcmp(p->tag,"Result")!=0)
        add_header(newHeaders,p->tag,p->val);
    }

    // write headers and cleanup
    n = newHeaders.GetSize();
    for (i =0; i<n; i++)
    {
        StringPair *p = (StringPair*)newHeaders.GetAt(i);
        ofile << "[" << p->tag << " \"" << p->val << "\"]" << endl;
        delete p;
    }
    newHeaders.RemoveAll();
    ofile << endl;

    // Write game moves.
    const int size = log.num_moves();
    char buf[80];
    *buf = '\0';
    char num[10];
    int last = 0;
    int num_image_size;
    for (i = 0; i < size; i++)
    {
        const Log_Entry &e = log[i];
        num_image_size = 0;
        *num = '\0';
        if (i % 2 == 0)
        {
            sprintf(num,"%d. ",(i/2)+1);
            num_image_size = strlen(num);
        }
        const int image_size = strlen(e.image());
        if (last + image_size + num_image_size + 1 >= PGN_MARGIN)
        {
            buf[last] = '\0';
            ofile << buf << endl;
            last = 0;
            *buf = '\0';
        }
        if (last != 0)
            strcat(buf," ");
        strcat(buf,num);
        strcat(buf,e.image());
        last = strlen(buf);
    }
    if (last + strlen(result) + 1 >= PGN_MARGIN)
    {
        ofile << buf << endl;
        last = 0;
        *buf = '\0';
    }
    strcat(buf," ");
    strcat(buf, result);
    ofile << buf << endl;
    if (!ofile) 
    {
#ifdef _WINDOWS
        ::MessageBox(NULL,"Error saving file!","Save Game",MB_OK);
#endif
        return FALSE;
    }
    return TRUE;
}

BOOL ChessIO::store_fen( ofstream &ofile, const Board &board)
{
    ofile << board << endl;
    return ofile.good();
}

EPD_Record * ChessIO::read_epd_record(istream &ifs, Board &board)
{
    // read FEN description
    EPD_Record *rec = new EPD_Record;
    ifs >> board;
    if (ifs.eof())
        return NULL;
    else if (!ifs.good())
    {
        rec->set_error("Bad EPD record: FEN board description missing or invalid");
        ifs.ignore(255,'\n');
        return rec;
    }
    // read EPD commands
    int c;
    while (ifs.good() && (c = ifs.get()) != EOF)
    {
        int saw_eoln = 0;
        while (isspace(c))
        {
            if (c == '\n' || c == '\r')
               saw_eoln++;
            c = ifs.get();
        }
        if (c == 0) break;
        if (saw_eoln)
        {
            ifs.putback(c);
            break;
        }
        // collect command
        char cmd[20], val[256];
        char *p = cmd;
        int count = 0;
        while (ifs.good() && !isspace(c) && count < 19)
        {
            *p++ = c;
            c = ifs.get();
            ++count;
        }
        *p = '\0';
        while (isspace(c))
        {
            c = ifs.get();
        }
        p = val;
        int quoted = (c == '"');
        count = 0;
        while (ifs.good() && 
               count < 255 &&
               (c != ';' || quoted))
        {
            *p++ = c;
            c = ifs.get();
            if (quoted && c == '"')
            {
               *p++ = c;
               quoted = 0;
               break;
            }
        }
        if (quoted)
        {
            rec->set_error("Missing end quote in EPD record");
            ifs.ignore(255,'\n');
            return rec;
        }
        *p = '\0';
        if (*cmd && *val)
            rec->add(cmd,val);
        while (!ifs.eof() && c != ';') c = ifs.get();
    }
    if (c == '\n') c = ifs.get();
    return rec;
}

#ifdef _WINDOWS
void ChessIO::execute_epd_record(EPD_Record *rec, Board &board)
{
    CString moveStr;
    if (rec->get_val("bm",moveStr))
    {
        // collect possible moves
        Move moves[10];
        char *buf = moveStr.GetBuffer(128);
        int move_count = 0;
        for (char *p = buf;*p;)
        {
            while (isspace(*p) && *p != '\0') ++p;
            if (*p == '\0')
               break;
            char tmp[10];
            int i = 0;
            char *q = tmp;
            while (!isspace(*p) && *p != '+' && *p != '\0' && i < 10)
            {
                *q++ = *p++;
                ++i;
            }
            *q = '\0'; 
            Move m = Notation::Value(board,board.Side(),tmp);
            if (!IsNull(m))
            {
                ASSERT(move_count < 10);
                moves[move_count++] = m;
            }
            if (*p == '+' || *p == '#') ++p;
        }
        moveStr.ReleaseBuffer();
        Search::Statistics stats;
        // use the internal ID_COMPUTE_ONLY message to perform
        // the search, w/o updating the board afterwards
        theView->SendMessage(WM_COMMAND,ID_COMPUTE_ONLY,0L);
        Move m = theView->get_last_move();
        int correct = 0;
        for (int i = 0; i < move_count; ++i)
        {
            if (moves[i] == m)
            {
               ++correct; break;
            }
        }    
        if (correct)
           theLog->write("+++ correct");
        else
           theLog->write("+++ error");
        theLog->write_eol();
    }
}

void ChessIO::annotate_epd_record(EPD_Record *rec, Board &board,
                                  ofstream &outfile)
{
    CString moveStr;
    char fenBuf[128];

    // Start the search (executes in another thread)
    theView->OnComputeMove();
    // Wait for the thread to complete ..
    theView->waitForSearch();
    const Search::Statistics &stats = theView->get_stats();
    Move m = theView->get_last_move();

    // write 1st 4 parts of a FEN record:
    ZeroMemory(fenBuf, 128);
	BoardIO::write_fen(board,fenBuf,FALSE);
    outfile << fenBuf << ' ';

    // Construct centipawn evaluation
    int val;
    if (stats.state == Search::Draw)
    {
       // The EPD standard requires that a draw position be
       // scored as 0
       val = 0;
    }
    else if (stats.value > 9000)
    {
       // EPD standard says a mate score is 32767
       val = 32767-(10000-stats.value);
    }
    else if (stats.value < -9000)
    {
       val = -32767+(10000+stats.value);
    }
    else
    {
       val = stats.value*100/64;
    }
    char valImage[20]; 
    sprintf(valImage,"%d",val);
    rec->add("ce",valImage);
    
    // write principal variation
    // EPD spec says we always need the pv tag even if there
    // are no moves.
    rec->add("pv",stats.best_line_image);

    // write predicted move if there is one
    if (!IsNull(stats.best_line[0]))
    {
       char movebuf[10];
       const char *p = stats.best_line_image;
       char *q = movebuf;
       while (*p && !isspace(*p))
          *q++=*p++;
       *q = '\0';
       rec->add("pm",movebuf);
    }
    // write node count and analysis time
    sprintf(valImage,"%ld",stats.num_nodes);
    rec->add("acn",valImage);
    sprintf(valImage,"%ld",stats.elapsed_time/100);
    rec->add("acs",valImage);

    // EPD standard says a "normalized" record has sorted operators
    rec->sort();

    // write out the EPD records
    const int n = rec->getSize();
    for (int i=0; i<n; i++)
    {
       CString key,val;
       rec->getData(i,key,val);
       outfile << key << ' ' << val << ';';
       if (i<n-1) outfile << ' ';           
    }
    outfile << endl;
}
#endif

